# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import deepcopy
from typing import Optional

import torch

import nncf
from nncf.common.factory import NNCFGraphFactory
from nncf.common.quantization.structs import QuantizationPreset
from nncf.data import Dataset
from nncf.parameters import BackupMode
from nncf.parameters import CompressWeightsMode
from nncf.parameters import ModelType
from nncf.parameters import QuantizationMode
from nncf.parameters import SensitivityMetric
from nncf.parameters import TargetDevice
from nncf.quantization.advanced_parameters import AdvancedCompressionParameters
from nncf.quantization.advanced_parameters import AdvancedQuantizationParameters
from nncf.quantization.algorithms.post_training.algorithm import PostTrainingQuantization
from nncf.quantization.algorithms.weight_compression.algorithm import WeightCompression
from nncf.quantization.quantize_model import warning_model_no_batchwise_support
from nncf.scopes import IgnoredScope
from nncf.torch.graph.operator_metatypes import OPERATIONS_OUTPUT_HAS_NO_BATCH_AXIS
from nncf.torch.model_creation import wrap_model

DEFAULT_RANGE_TYPE = "mean_min_max"


def quantize_impl(
    model: torch.nn.Module,
    calibration_dataset: Dataset,
    mode: Optional[QuantizationMode] = None,
    preset: Optional[QuantizationPreset] = None,
    target_device: TargetDevice = TargetDevice.ANY,
    subset_size: int = 300,
    fast_bias_correction: bool = True,
    model_type: Optional[ModelType] = None,
    ignored_scope: Optional[IgnoredScope] = None,
    advanced_parameters: Optional[AdvancedQuantizationParameters] = None,
) -> torch.nn.Module:
    """
    Implementation of the `quantize()` method for the PyTorch backend.
    """
    if fast_bias_correction is False:
        raise ValueError(f"fast_bias_correction={fast_bias_correction} is not supported")
    if target_device == TargetDevice.CPU_SPR:
        raise nncf.InternalError("target_device == CPU_SPR is not supported")
    if mode is not None:
        raise ValueError(f"mode={mode} is not supported")

    copied_model = deepcopy(model)

    example_input = next(iter(calibration_dataset.get_inference_data()))
    nncf_network = wrap_model(copied_model.eval(), example_input, trace_parameters=True)

    quantization_algorithm = PostTrainingQuantization(
        preset=preset,
        target_device=target_device,
        subset_size=subset_size,
        fast_bias_correction=fast_bias_correction,
        model_type=model_type,
        ignored_scope=ignored_scope,
        advanced_parameters=advanced_parameters,
    )
    graph = nncf_network.nncf.get_graph()
    warning_model_no_batchwise_support(graph, advanced_parameters, model_type, OPERATIONS_OUTPUT_HAS_NO_BATCH_AXIS)
    quantized_model = quantization_algorithm.apply(nncf_network, graph, dataset=calibration_dataset)

    quantized_model.nncf.disable_dynamic_graph_building()

    return quantized_model


def compress_weights_impl(
    model: torch.nn.Module,
    dataset: Dataset,
    mode: CompressWeightsMode,
    ratio: float,
    group_size: int,
    ignored_scope: IgnoredScope,
    all_layers: bool,
    sensitivity_metric: SensitivityMetric,
    awq: bool,
    subset_size: int,
    scale_estimation: bool,
    gptq: bool,
    lora_correction: bool,
    backup_mode: BackupMode,
    advanced_parameters: Optional[AdvancedCompressionParameters] = None,
) -> torch.nn.Module:
    """
    Implementation of the `compress_weights()` method for the PyTorch backend.
    """

    compression_algorithm = WeightCompression(
        mode,
        ratio,
        group_size,
        ignored_scope,
        all_layers,
        sensitivity_metric,
        awq,
        subset_size,
        scale_estimation,
        gptq,
        lora_correction,
        backup_mode,
        advanced_parameters,
    )
    graph = NNCFGraphFactory.create(model)
    return compression_algorithm.apply(model, graph, dataset=dataset)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path
import subprocess

import torch

import nncf
from nncf import nncf_logger
from nncf.definitions import NNCF_PACKAGE_ROOT_DIR
from nncf.torch.extensions import EXTENSIONS
from nncf.torch.extensions import CudaNotAvailableStub
from nncf.torch.extensions import ExtensionLoader
from nncf.torch.extensions import ExtensionLoaderTimeoutException
from nncf.torch.extensions import ExtensionNamespace
from nncf.torch.extensions import ExtensionsType
from nncf.torch.quantization.reference import ReferenceQuantizedFunctions

BASE_EXT_DIR = os.path.join(NNCF_PACKAGE_ROOT_DIR, "torch/extensions/src/quantization")

EXT_INCLUDE_DIRS = [
    os.path.join(NNCF_PACKAGE_ROOT_DIR, "torch/extensions/include"),
]

CPU_EXT_SRC_LIST = [
    os.path.join(BASE_EXT_DIR, "cpu/functions_cpu.cpp"),
    os.path.join(NNCF_PACKAGE_ROOT_DIR, "torch/extensions/src/common/cpu/tensor_funcs.cpp"),
]

CUDA_EXT_SRC_LIST = [
    os.path.join(BASE_EXT_DIR, "cuda/functions_cuda.cpp"),
    os.path.join(BASE_EXT_DIR, "cuda/functions_cuda_impl.cu"),
]


@EXTENSIONS.register()
class QuantizedFunctionsCPULoader(ExtensionLoader):
    @classmethod
    def extension_type(cls):
        return ExtensionsType.CPU

    @classmethod
    def name(cls) -> str:
        return "quantized_functions_cpu"

    @classmethod
    def load(cls):
        try:
            retval = torch.utils.cpp_extension.load(
                cls.name(),
                CPU_EXT_SRC_LIST,
                extra_include_paths=EXT_INCLUDE_DIRS,
                build_directory=cls.get_build_dir(),
                verbose=False,
            )
        except ExtensionLoaderTimeoutException as e:
            raise e
        except Exception as e:
            nncf_logger.warning(
                f"Could not compile CPU quantization extensions. "
                f"Falling back on torch native operations - "
                f"CPU quantization fine-tuning may be slower than expected.\n"
                f"Reason: {str(e)}"
            )
            retval = ReferenceQuantizedFunctions
        return retval


@EXTENSIONS.register()
class QuantizedFunctionsCUDALoader(ExtensionLoader):
    @classmethod
    def extension_type(cls):
        return ExtensionsType.CUDA

    @classmethod
    def load(cls):
        try:
            return torch.utils.cpp_extension.load(
                cls.name(),
                CUDA_EXT_SRC_LIST,
                extra_include_paths=EXT_INCLUDE_DIRS,
                build_directory=cls.get_build_dir(),
                verbose=False,
            )
        except ExtensionLoaderTimeoutException as e:
            raise e
        except (subprocess.CalledProcessError, OSError, RuntimeError) as e:
            assert torch.cuda.is_available()
            raise nncf.InstallationError(
                "CUDA is available for PyTorch, but NNCF could not compile "
                "GPU quantization extensions. Make sure that you have installed CUDA development "
                "tools (see https://docs.nvidia.com/cuda/cuda-installation-guide-linux/index.html for "
                "guidance) and that 'nvcc' is available on your system's PATH variable.\n"
            ) from e

    @classmethod
    def name(cls) -> str:
        return "quantized_functions_cuda"


QuantizedFunctionsCPU = ExtensionNamespace(QuantizedFunctionsCPULoader())

if torch.cuda.is_available():
    QuantizedFunctionsCUDA = ExtensionNamespace(QuantizedFunctionsCUDALoader())
else:
    QuantizedFunctionsCUDA = CudaNotAvailableStub()



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import deque

import torch.fx

from nncf.common.factory import ModelTransformerFactory
from nncf.common.graph.definitions import NNCFGraphNodeType
from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.graph.transformations.layout import TransformationLayout
from nncf.experimental.torch.fx.commands import FXApplyTransformationCommand
from nncf.experimental.torch.fx.transformations import node_removal_transformation_builder
from nncf.torch.graph.operator_metatypes import QUANTIZE_NODE_TYPES
from nncf.torch.graph.transformations.commands import PTTargetPoint


def remove_fq_from_inputs(model: torch.fx.GraphModule, graph: NNCFGraph) -> torch.fx.GraphModule:
    """
    This method removes the activation Fake Quantize nodes from the model.
    It's needed for the further bias shift calculation that relates on quantized weights.

    :param model: ov.Model instance.
    :param graph: NNCFGraph instance.
    :return: ov.Model instance without activation Fake Quantize nodes.
    """
    transformation_layout = TransformationLayout()
    model_transformer = ModelTransformerFactory.create(model)

    seen_nodes = []
    nodes_queue = deque(graph.get_input_nodes())
    while nodes_queue:
        current_node = nodes_queue.popleft()
        current_node_name = current_node.node_name

        if current_node_name in seen_nodes:
            continue

        seen_nodes.append(current_node_name)
        if current_node.node_type in QUANTIZE_NODE_TYPES:
            transformation = node_removal_transformation_builder(current_node, input_port_id=0)
            transformation_layout.register(FXApplyTransformationCommand(transformation))
        nodes_queue.extend(graph.get_next_nodes(current_node))

    return model_transformer.transform(transformation_layout)


_TARGET_TYPE_TO_FX_INS_TYPE_MAP = {
    TargetType.PRE_LAYER_OPERATION: TargetType.OPERATOR_PRE_HOOK,
    TargetType.POST_LAYER_OPERATION: TargetType.OPERATOR_POST_HOOK,
}


def get_target_point(target_type: TargetType, target_node_name: str, port_id: int) -> PTTargetPoint:
    """
    Creates torch-specific target point.

    :param target_type: Target point target type.
    :param target_node_name: Target node name to use in the target point.
    :param port_id: Target port id.
    :return: Torch-specific target point.
    """
    if NNCFGraphNodeType.INPUT_NODE in target_node_name or target_type == TargetType.POST_LAYER_OPERATION:
        port_id = None
    if target_type in _TARGET_TYPE_TO_FX_INS_TYPE_MAP:
        target_type = _TARGET_TYPE_TO_FX_INS_TYPE_MAP[target_type]
    return PTTargetPoint(target_type, target_node_name, input_port_id=port_id)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from nncf.tensor import Tensor
from nncf.tensor import TensorDataType
from nncf.tensor import functions as fns


def mean_per_channel(x: Tensor, axis: int, dtype: Optional[TensorDataType] = None) -> Tensor:
    """
    Computes the mean of elements across given channel dimension of Tensor.

    :param x: Tensor to reduce.
    :param axis: The channel dimensions to reduce.
    :param dtype: Type to use in computing the mean.
    :return: Reduced Tensor.
    """
    if len(x.shape) < 3:
        return fns.mean(x, axis=0, dtype=dtype)

    pos_axis = axis + x.ndim if axis < 0 else axis
    if pos_axis < 0 or pos_axis >= x.ndim:
        raise ValueError(f"axis {axis} is out of bounds for array of dimension {x.ndim}")
    axis = tuple(i for i in range(x.ndim) if i != pos_axis)
    return fns.mean(x, axis=axis, dtype=dtype)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from pathlib import Path
from typing import Dict, List, Tuple

from nncf.api.compression import TModel
from nncf.common.factory import StatisticsAggregatorFactory
from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.common.tensor_statistics.aggregator import StatisticsAggregator
from nncf.data import Dataset
from nncf.parameters import SensitivityMetric
from nncf.quantization.algorithms.weight_compression.algorithm import WeightCompression
from nncf.quantization.algorithms.weight_compression.algorithm import get_weight_compression_configuration
from nncf.quantization.algorithms.weight_compression.mixed_precision import MIXED_PRECISION_CRITERIA


def register_statistics_for_algorithm(
    aggregator: StatisticsAggregator,
    model: TModel,
    graph: NNCFGraph,
    compression_algo: WeightCompression,
    matmul_input_to_output_nodes_map: Dict[Tuple[NNCFNode, int], List[NNCFNode]],
) -> None:
    """
    Registers the statistics required for the given compression algorithm.

    :param aggregator: Aggregator to register statistics.
    :param model: Model being analyzed.
    :param graph: Model's computational graph.
    :param compression_algo: WeightCompression algorithm instance.
    :param matmul_input_to_output_nodes_map: A dictionary mapping from a tuple of (activation node, port ID)
    to a list of MatMul nodes that accept the activation as input.
    """
    statistic_points = compression_algo.get_statistic_points(model, graph, matmul_input_to_output_nodes_map.keys())
    aggregator.register_statistic_points(statistic_points)


def _register_mixed_precision(
    aggregator: StatisticsAggregator,
    model: TModel,
    graph: NNCFGraph,
    matmul_input_to_output_nodes_map: Dict[Tuple[NNCFNode, int], List[NNCFNode]],
    subset_size: int,
) -> None:
    """
    Registers statistics for mixed precision compression algorithm.

    :param aggregator: Aggregator to register statistics.
    :param model: Model being analyzed.
    :param graph: Model's computational graph.
    :param matmul_input_to_output_nodes_map: A dictionary mapping from a tuple of (activation node, port ID)
    to a list of MatMul nodes that accept the activation as input.
    :param subset_size: Size of dataset subset for statistics.
    """
    sensitivities = [
        SensitivityMetric.HESSIAN_INPUT_ACTIVATION,
        SensitivityMetric.MEAN_ACTIVATION_VARIANCE,
        SensitivityMetric.MAX_ACTIVATION_VARIANCE,
        SensitivityMetric.MEAN_ACTIVATION_MAGNITUDE,
    ]

    for sensitivity in sensitivities:
        criterion_cls = MIXED_PRECISION_CRITERIA.get(sensitivity)
        mixed_prec_algo = criterion_cls(None, None, subset_size)
        statistic_points = mixed_prec_algo.get_statistic_points(model, graph, matmul_input_to_output_nodes_map.keys())
        aggregator.register_statistic_points(statistic_points)


def register_all_statistics(
    aggregator: StatisticsAggregator,
    model: TModel,
    graph: NNCFGraph,
    subset_size: int,
    compression_algo: WeightCompression,
    enable_mixed_precision: bool = True,
) -> None:
    """
    Registers all required statistics for the model compression.

    :param aggregator: Aggregator to register statistics.
    :param model: Model being analyzed.
    :param graph: Model's computational graph.
    :param compression_algo: WeightCompression algorithm instance.
    :param enable_mixed_precision: Whether to enable mixed precision statistics.
    """
    _, matmul_input_to_output_nodes_map = compression_algo.get_compression_nodes_info(graph)

    register_statistics_for_algorithm(aggregator, model, graph, compression_algo, matmul_input_to_output_nodes_map)

    if enable_mixed_precision:
        _register_mixed_precision(aggregator, model, graph, matmul_input_to_output_nodes_map, subset_size)


def cache_weight_compression_statistics(
    model: TModel, graph: NNCFGraph, dataset: Dataset, subset_size: int, statistics_path: Path
) -> None:
    """
    Caches compression statistics for a given model and dataset.

    :param model: Model being analyzed.
    :param graph: Model's computational graph.
    :param dataset: Dataset to analyze model statistics.
    :param subset_size: Size of dataset subset for statistics.
    :param statistics_path: Path to save cached statistics.
    """
    config = get_weight_compression_configuration(awq=True, scale_estimation=True, lora_correction=True)
    compression_algo = WeightCompression(**config, subset_size=subset_size)
    compression_algo.set_backend_entity(model)
    aggregator = StatisticsAggregatorFactory.create(model, dataset)
    register_all_statistics(aggregator, model, graph, subset_size, compression_algo)
    aggregator.collect_statistics(model, graph)
    aggregator.dump_statistics(statistics_path)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Tuple, Union

from nncf.common.utils.caching import disable_results_caching
from nncf.openvino.optimized_functions.models import OV_MODEL_CACHE
from nncf.openvino.optimized_functions.models import OVModelParameters
from nncf.openvino.optimized_functions.models import get_astype_model
from nncf.openvino.optimized_functions.models import get_compress_decompress_weight_model
from nncf.openvino.optimized_functions.models import get_compress_weight_model
from nncf.quantization.algorithms.weight_compression.config import WeightCompressionConfig
from nncf.quantization.algorithms.weight_compression.weight_lowering import reshape_weight_for_grouped_quantization
from nncf.tensor import Tensor
from nncf.tensor import TensorBackend
from nncf.tensor import TensorDataType

ReductionAxes = Union[int, Tuple[int, ...]]


def do_int_quantization(
    weight: Tensor,
    config: WeightCompressionConfig,
    reduction_axes: Optional[ReductionAxes] = None,
    precomputed_scale: Tensor = None,
    precomputed_zero_point: Tensor = None,
    **kwargs,
) -> Tuple[Tensor, Tensor, Tensor]:
    """
    Quantizes the given weight tensor.

    :param weight: The weight tensor to quantize.
    :param config: The weight compression configuration.
    :param reduction_axes: Axes along which to reduce (collect) statistics (e.g., min, max). Not required if
        precomputed scale (and zero point) are provided.
    :param precomputed_scale: Optional precomputed scale tensor.
    :param precomputed_zero_point: Optional precomputed zero point tensor.
    :return: A tuple containing the compressed weights, scale, and zero point tensors.
    """
    weight_shape = weight.shape
    scale_shape = None if precomputed_scale is None else precomputed_scale.shape
    zero_point_shape = None if precomputed_zero_point is None else precomputed_zero_point.shape

    ov_model_params = OVModelParameters(
        dynamic_shapes=kwargs.get("dynamic_shapes") is True,
        convertable_division=kwargs.get("convertable_division") is True,
    )
    ov_model_params.input_dtypes["weight"] = weight.dtype
    if precomputed_scale is not None:
        ov_model_params.input_dtypes["scale"] = precomputed_scale.dtype
    if precomputed_zero_point is not None:
        ov_model_params.input_dtypes["zero_point"] = precomputed_zero_point.dtype
    if config.num_bits == 4 and weight.backend == TensorBackend.ov:
        # Return ov tensors in target precision to seamlessly insert them into openvino model later
        ov_model_params.return_ov_tensors = True
        compressed_weight_dtype = TensorDataType.uint4 if config.is_asym_mode else TensorDataType.int4
        ov_model_params.output_dtypes.update(
            {"compressed_weight": compressed_weight_dtype, "zero_point": compressed_weight_dtype}
        )

    model = get_compress_weight_model(
        ov_model_params,
        config,
        weight_shape,
        scale_shape,
        zero_point_shape,
        reduction_axes,
    )

    if precomputed_scale is None:
        # weight -> compressed_weight, scale, (zero_point)
        results = model([weight])
        if config.is_asym_mode:
            compressed_weight, scale, zero_point = results
        else:
            compressed_weight, scale = results
            zero_point = None

        # Scale is always in fp32 so there is no need to store it in ov.Tensor
        if scale.backend == TensorBackend.ov:
            scale = scale.as_numpy_tensor()
    else:
        # weight, scale, (zero_point) -> compressed_weight
        inputs = (
            [weight, precomputed_scale]
            if precomputed_zero_point is None
            else [weight, precomputed_scale, precomputed_zero_point]
        )
        compressed_weight = model(inputs)[0]
        scale, zero_point = precomputed_scale, precomputed_zero_point

    return compressed_weight, scale, zero_point


def quantize_dequantize_weight(
    weight: Tensor,
    config: WeightCompressionConfig,
    reduction_axes: Optional[ReductionAxes] = None,
    precomputed_scale: Optional[Tensor] = None,
    precomputed_zero_point: Optional[Tensor] = None,
    return_compressed_weight: Optional[bool] = False,
    **kwargs,
) -> Union[Tensor, Tuple[Tensor, Tensor, Tensor, Tensor]]:
    """
    Quantizes the given weight tensor and then dequantizes it back to obtain float32 values.

    :param weight: The weight tensor to quantize-dequantize.
    :param config: Compression configuration.
    :param reduction_axes: Axes along which to reduce (collect) statistics (e.g., min, max). Not required if
        precomputed scale (and zero point) are provided.
    :param precomputed_scale: Optional precomputed scale tensor.
    :param precomputed_zero_point: Optional precomputed zero point tensor.
    :param return_compressed_weight: If True, besides decompressed weight will also return compressed weight, scale,
        (and zero point).
    :return: Dequantized weight tensor or a tuple containing the decompressed weight, compressed weight, scale,
        (and zero point).
    """
    # When reduction axes are not provided, assuming that the weights are already reshaped
    if config.group_size != -1 and reduction_axes is not None:
        # weights are reshaped from [a1, r, a2] to [a1, r//gs, gs, a2]
        weight, reduction_axes = reshape_weight_for_grouped_quantization(weight, reduction_axes, config.group_size)

    weight_shape = weight.shape
    scale_shape = precomputed_scale.shape if precomputed_scale is not None else None
    zero_point_shape = precomputed_zero_point.shape if precomputed_zero_point is not None else None

    ov_model_params = OVModelParameters(
        dynamic_shapes=kwargs.get("dynamic_shapes") is True,
        convertable_division=kwargs.get("convertable_division") is True,
    )
    ov_model_params.input_dtypes["weight"] = weight.dtype
    if precomputed_scale is not None:
        ov_model_params.input_dtypes["scale"] = precomputed_scale.dtype
    if precomputed_zero_point is not None:
        ov_model_params.input_dtypes["zero_point"] = precomputed_zero_point.dtype

    model = get_compress_decompress_weight_model(
        ov_model_params, config, weight_shape, scale_shape, zero_point_shape, reduction_axes, return_compressed_weight
    )

    inputs = [weight]
    if precomputed_scale is not None:
        inputs.append(precomputed_scale)
    if precomputed_zero_point is not None:
        inputs.append(precomputed_zero_point)

    compressed_weight, scale, zero_point = None, precomputed_scale, precomputed_zero_point
    results = model(inputs)
    if len(results) == 1:
        decompressed_weight = results[0]
    elif len(results) == 2:
        decompressed_weight, compressed_weight = results
    elif len(results) == 3:
        decompressed_weight, compressed_weight, scale = results
    else:
        decompressed_weight, compressed_weight, scale, zero_point = results
    if return_compressed_weight:
        return decompressed_weight, compressed_weight, scale, zero_point
    else:
        return decompressed_weight


def astype(a: Tensor, dtype: TensorDataType) -> Tensor:
    """
    Converts the given tensor to the specified data type. Allows to convert between u4, i4, bf16 data types which are
    not natively supported by numpy. These data types are represented as the following data types in numpy:
        - bfloat16 -> np.float16,
        - uint4 -> uint8,
        - int4 -> int8.
    :param a: Tensor to change data type for.
    :param dtype: Data type to convert to.
    :return: Tensor with the specified data type.
    """
    ov_model_params = OVModelParameters(
        input_dtypes={"input": a.dtype},
        output_dtypes={"output": dtype},
        release_memory=False,
        return_ov_tensors=True,
    )
    with disable_results_caching(OV_MODEL_CACHE):
        model = get_astype_model(ov_model_params, tuple(a.shape))
    return model([Tensor(a)])[0]



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import defaultdict
from typing import Dict, Iterator, List, Optional, Union

import numpy as np
import onnx
from onnx import numpy_helper

import nncf


def get_name_to_node_map(model: onnx.ModelProto) -> Dict[str, onnx.NodeProto]:
    """
    Returns mapping from node name to the node.

    :param model: Model from mapping is built.
    :return: Mapping.
    """
    return {node.name: node for node in model.graph.node}


def get_edge_info_mapping(model: onnx.ModelProto) -> Dict[str, onnx.ValueInfoProto]:
    """
    Retuns mapping from edge name to the edge info.

    :param model: Model from mapping is built.
    :return: Mapping.
    """
    return {
        tensor.name: tensor
        for tensor in (*model.graph.value_info, *model.graph.input, *model.graph.output, *model.graph.initializer)
    }


def get_children_node_mapping(model: onnx.ModelProto) -> Dict[str, List[onnx.NodeProto]]:
    """
    Returns a mapping from edge name to nodes which consume this edge as an input.

    :param model: ONNX model.
    :return: Mapping from edge name to nodes which consume this edge as an input.
    """
    output = defaultdict(list)
    for node in model.graph.node:
        for edge in node.input:
            output[edge].append(node)
    return output


def get_parents_node_mapping(model: onnx.ModelProto) -> Dict[str, onnx.NodeProto]:
    """
    Returns a mapping from edge name to node which outputs this edge.

    :param model: ONNX model.
    :return: Mapping from edge name to node which outputs this edge.
    """
    output = {}
    for node in model.graph.node:
        for edge in node.output:
            output[edge] = node
    return output


def get_model_inputs(model: onnx.ModelProto) -> List[onnx.ValueInfoProto]:
    """
    Returns all model inputs.

    :param model: ONNX model.
    :return: Model Inputs.
    """
    inputs = []
    input_all = [node.name for node in model.graph.input]
    input_initializer = [node.name for node in model.graph.initializer]
    net_feed_input = list(set(input_all) - set(input_initializer))
    for node in model.graph.input:
        if node.name in net_feed_input:
            inputs.append(node)
    return inputs


def get_input_port_id_for_node_after_input(input_name: str, to_node: onnx.NodeProto) -> int:
    """
    Returns input_port_id for 'to_node' connected with the model input with the name 'input_name'.

    :param input_name: Name of the ONNX model Input.
    :param to_node: Node, which has input edge with 'input_name' name.
    :return: input port number for 'to_node', which is connected to 'input_name'.
    """
    for input_port_id, port in enumerate(to_node.input):
        if port == input_name:
            return input_port_id
    raise nncf.ValidationError(f"The node {to_node} does not have input edge with the name {input_name}")


def get_output_port_id_for_node_before_output(output_name: str, from_node: onnx.NodeProto) -> int:
    """
    Returns output_port_id for 'from_node' connected with the model output with the name 'output_name'.

    :param output_name: Name of the ONNX model Output.
    :param from_node: Node, which has output edge with 'output_name' name.
    :return: output port number for 'from_node', which is connected to 'output_name'.
    """
    for output_port_id, port in enumerate(from_node.output):
        if port == output_name:
            return output_port_id
    raise nncf.ValidationError(f"The node {from_node} does not have output edge with the name {output_name}")


def get_node_index(model: onnx.ModelProto, node_name: str) -> Optional[int]:
    """
    Returns the node index in the model.

    :param model: ONNX model.
    :param node_name: Name of the node.
    :return: Node index, -1 if there is no such node.
    """
    for i, node in enumerate(model.graph.node):
        if node.name == node_name:
            return i
    return None


def _get_all_tensors(model: onnx.ModelProto) -> Iterator[onnx.TensorProto]:
    """
    Iterate over all tensors of ONNX model.

    :param model: ONNX model.
    :yield: tensors of ONNX model.
    """
    for initializer in model.graph.initializer:
        yield initializer
    for node in model.graph.node:
        for attribute in node.attribute:
            if attribute.HasField("t"):
                yield attribute.t
            yield from attribute.tensors


def has_tensor(model: onnx.ModelProto, tensor_name: str) -> bool:
    """
    Returns True whether the model has the tensor with the name equals to tensor_name.

    :param model: ONNX model.
    :param tensor_name: Name of the tensor.
    :return: True if the model has such tensor, False - otherwise.
    """
    for tensor in _get_all_tensors(model):
        if tensor.name == tensor_name:
            return True
    return False


def get_tensor(model: onnx.ModelProto, tensor_name: str) -> onnx.TensorProto:
    """
    Returns a tensor with the name 'tensor_name'.

    :param model: ONNX model.
    :param tensor_name: Name of the tensor.
    :return: The Initializer.
    """
    for tensor in _get_all_tensors(model):
        if tensor.name == tensor_name:
            return tensor
    raise nncf.ValidationError("There is no tensor with the name {}".format(tensor_name))


def get_tensor_value(model: onnx.ModelProto, tensor_name: str) -> np.ndarray:
    """
    Returns tensor value of a tensor with the name 'tensor_name'.

    :param model: ONNX model.
    :param tensor_name: Name of the tensor.
    :return: The value of the tensor.
    """
    return numpy_helper.to_array(get_tensor(model, tensor_name))


def get_edge_shape(edge: Union[onnx.ValueInfoProto, onnx.TensorProto]) -> List[int]:
    """
    Returns edge shape.

    :param edge: The edge.
    :return: Shape of the Tensor.
    """
    if isinstance(edge, onnx.TensorProto):
        return list(edge.dims)
    tensor_type = edge.type.tensor_type
    shape = []
    if tensor_type.HasField("shape"):
        for d in tensor_type.shape.dim:
            if d.HasField("dim_value"):
                dim_value = d.dim_value
                if isinstance(dim_value, int):
                    shape.append(dim_value)
                else:
                    return shape
            elif d.HasField("dim_param"):
                # flexible shape  make manually -1
                shape.append(-1)
            else:
                return shape
    return shape


def get_edge_dtype(edge: Union[onnx.ValueInfoProto, onnx.TensorProto]) -> int:
    """
    Returns the data type of the edge.

    :param edge: The edge.
    :return: Data type of the edge.
    """
    if isinstance(edge, onnx.ValueInfoProto):
        return edge.type.tensor_type.elem_type
    return edge.data_type


def get_parent(
    node: onnx.NodeProto,
    port_id: int,
    parents_node_mapping: Dict[str, onnx.NodeProto],
) -> Optional[onnx.NodeProto]:
    """
    Returns parents of the node. If there is no parent node, returns None.

    :param node: The child node.
    :param port_id: Input port id on which the parent is seeked.
    :param edge_node_mapping: Mapping describing start and consumed nodes of the edges.
    :return: Parent node.
    """
    if port_id < len(node.input):
        return parents_node_mapping.get(node.input[port_id])
    return None


def get_children(node: onnx.NodeProto, children_node_mapping: Dict[str, List[onnx.NodeProto]]) -> List[onnx.NodeProto]:
    """
    Returns children of the node.

    :param node: The parent node.
    :param edge_node_mapping: Mapping describing start and consumed nodes of the edges.
    :return: All children nodes.
    """
    output = []
    for node_edge in node.output:
        output.extend(children_node_mapping[node_edge])
    return output


def is_node_has_shared_weight(
    node: onnx.NodeProto,
    weight_port_id: int,
    children_node_mapping: Dict[str, List[onnx.NodeProto]],
) -> bool:
    """
    Returns whether the node share a weight.

    :param node: Node.
    :param weight_port_id: Port id on which there is a weight.
    :param edge_node_mapping: Mapping describing start and consumed nodes of the edges.
    :return: True whether node shares a weight - otherwise False.
    """
    weight_tensor_edge = node.input[weight_port_id]
    nodes = children_node_mapping[weight_tensor_edge]
    return len(nodes) > 1



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Callable, Generator, Iterable, Iterator, List, Optional, cast

from nncf.common.utils.api_marker import api


@api(canonical_alias="nncf.Dataset")
class Dataset:
    """
    Wrapper for passing custom user datasets into NNCF algorithms.

    This class defines the interface by which compression algorithms
    retrieve data items from the passed data source object. These data items are used
    for different purposes, for example, model inference and model validation, based
    on the choice of the exact compression algorithm.

    If the data item has been returned from the data source per iteration and it cannot be
    used as input for model inference, the transformation function is used to extract the
    model's input from this data item. For example, in supervised learning, the data item
    usually contains both examples and labels. So transformation function should extract
    the examples from the data item.

    :param data_source: The iterable object serving as the source of data items.
    :param transform_func: The function that is used to extract the model's input
        from the data item. The data item here is the data item that is returned from
        the data source per iteration. This function should be passed when
        the data item cannot be directly used as model's input. If this is not specified, then the data item
        will be passed into the model as-is.
    """

    def __init__(self, data_source: Iterable[Any], transform_func: Optional[Callable[..., Any]] = None):
        self._data_source = data_source
        self._transform_func = transform_func

    def get_data(self, indices: Optional[List[int]] = None) -> Iterable[Any]:
        """
        Returns the iterable object that contains selected data items from the data source as-is.

        :param indices: The zero-based indices of data items that should be selected from
            the data source. The indices should be sorted in ascending order. If indices are
            not passed all data items are selected from the data source.
        :return: The iterable object that contains selected data items from the data source as-is.
        """
        return DataProvider(self._data_source, None, indices)

    def get_inference_data(self, indices: Optional[List[int]] = None) -> Iterable[Any]:
        """
        Returns the iterable object that contains selected data items from the data source, for which
        the transformation function was applied. The item, which was returned per iteration from this
        iterable, can be used as the model's input for model inference.

        :param indices: The zero-based indices of data items that should be selected from
            the data source. The indices should be sorted in ascending order. If indices are
            not passed all data items are selected from the data source.
        :return: The iterable object that contains selected data items from the data source, for which
            the transformation function was applied.
        """
        return DataProvider(self._data_source, self._transform_func, indices)

    def get_length(self) -> Optional[int]:
        """
        Tries to fetch length of the underlying dataset.
        :return: The length of the data_source if __len__() is implemented for it, and None otherwise.
        """
        if hasattr(self._data_source, "__len__"):
            return cast(int, self._data_source.__len__())
        return None

    def get_batch_size(self) -> Optional[int]:
        """
        Tries to fetch batch size of the underlying dataset.
        :return: The value of batch_size or _batch_size attributes of the data_source if exist, and None otherwise.
        """
        if hasattr(self._data_source, "batch_size"):  # Torch dataloader
            return cast(int, self._data_source.batch_size)
        if hasattr(self._data_source, "_batch_size"):  # TF dataloader
            return cast(int, self._data_source._batch_size)
        return None


class DataProvider:
    def __init__(
        self,
        data_source: Iterable[Any],
        transform_func: Optional[Callable[..., Any]],
        indices: Optional[List[int]] = None,
    ):
        self._data_source = data_source
        if transform_func is None:
            self._transform_func = lambda x: x
        else:
            self._transform_func = transform_func
        self._indices = indices

    def __iter__(self) -> Iterator[Any]:
        if self._indices is None:
            return map(self._transform_func, self._data_source)

        if hasattr(self._data_source, "__getitem__"):
            return DataProvider._get_iterator_for_map_style(self._data_source, self._transform_func, self._indices)

        return DataProvider._get_iterator_for_iter(self._data_source, self._transform_func, sorted(self._indices))

    @staticmethod
    def _get_iterator_for_map_style(
        data_source: Iterable[Any], transform_func: Callable[..., Any], indices: List[int]
    ) -> Generator[Any, None, None]:
        for index in indices:
            yield transform_func(data_source[index])  # type: ignore[index]

    @staticmethod
    def _get_iterator_for_iter(
        data_source: Iterable[Any], transform_func: Callable[..., Any], indices: List[int]
    ) -> Generator[Any, None, None]:
        pos = 0
        num_indices = len(indices)
        for idx, data_item in enumerate(data_source):
            if pos == num_indices:
                # All specified data items were selected.
                break
            if idx == indices[pos]:
                pos = pos + 1
                yield transform_func(data_item)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf

from nncf.common.statistics import NNCFStatistics
from nncf.tensorflow.callbacks.statistics_callback import StatisticsCallback


class UpdateMask(tf.keras.callbacks.Callback):
    def __init__(self, scheduler):
        super().__init__()
        self._scheduler = scheduler

    def on_train_batch_begin(self, batch, logs=None):
        self._scheduler.step()

    def on_epoch_begin(self, epoch, logs=None):
        self._scheduler.epoch_step(epoch)


class SparsityStatisticsCallback(StatisticsCallback):
    """
    Callback for logging sparsity compression statistics to tensorboard and stdout.
    """

    def _prepare_for_tensorboard(self, stats: NNCFStatistics):
        base_prefix = "2.compression/statistics"
        detailed_prefix = "3.compression_details/statistics"

        if stats.magnitude_sparsity:
            stats = stats.magnitude_sparsity
        else:
            stats = stats.rb_sparsity

        ms = stats.model_statistics
        tensorboard_stats = {
            f"{base_prefix}/sparsity_level_for_model": ms.sparsity_level,
            f"{base_prefix}/sparsity_level_for_sparsified_layers": ms.sparsity_level_for_layers,
            f"{base_prefix}/target_sparsity_level": stats.target_sparsity_level,
        }

        for ls in ms.sparsified_layers_summary:
            layer_name, sparsity_level = ls.name, ls.sparsity_level
            tensorboard_stats[f"{detailed_prefix}/{layer_name}/sparsity_level"] = sparsity_level

        return tensorboard_stats



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
from typing import Any, Callable, Dict, List, Optional

import torch.fx
import torch.utils._pytree as pytree

aten = torch.ops.aten


def _replace_node_with_constant(
    gm: torch.fx.GraphModule,
    node: torch.fx.Node,
    value: torch.Tensor,
) -> None:
    """
    Replaces given node to a constant with given value.

    :param gm: GraphModule instance.
    :param node: A node to replace with a constant.
    :param value: Value to use as a constant instead of the given node.
    """
    g = gm.graph

    if not hasattr(gm, "_frozen_param_count"):
        gm._frozen_param_count = 0  # type: ignore[assignment]
    i = gm._frozen_param_count

    while True:
        qualname = f"_frozen_param{i}"
        if not hasattr(gm, qualname):
            break
        i += 1

    gm._frozen_param_count = i + 1

    with g.inserting_before(node):
        new_input_node = g.create_node("get_attr", qualname, (), {})
        node.replace_all_uses_with(new_input_node)
        new_input_node.meta.update(node.meta)
        g.erase_node(node)

    # needed to suppress `does not reference an nn.Module, nn.Parameter, or buffer` warning
    gm.register_buffer(qualname, value)
    setattr(gm, qualname, value)


def _is_const_source(node: torch.fx.Node) -> bool:
    """
    Return True if node is a constant node.

    :param node: Given node.
    :return: True if node is a constant node.
    """
    return node.op == "get_attr"


class ConstantFolder(torch.fx.Interpreter):
    """
    Finds constant subraphs and place it in node_replacement attribute.
    """

    def __init__(
        self,
        gm: torch.fx.GraphModule,
    ) -> None:
        super().__init__(gm)
        self.node_replacements: Dict[torch.fx.Node, Any] = {}
        self.replaced_uses: Dict[torch.fx.Node, int] = collections.Counter()
        self.unknown_value = object()

        # overwrite this to deallocate env values if their only remaining use
        # is the output
        self.user_to_last_uses = self.node_to_last_non_output_use()

    def _deduce_value(self, node: torch.fx.Node) -> Any:
        return super().run_node(node)

    def is_impure(self, node: torch.fx.node.Node) -> bool:
        def is_woq_int8_pattern(node: torch.fx.node.Node) -> bool:
            return (
                node.target == torch.ops.prims.convert_element_type.default  # type: ignore[return-value]
                and isinstance(node.args[0], torch.fx.Node)
                and "val" in node.args[0].meta
                and node.args[0].meta["val"].dtype == torch.int8  # type: ignore[union-attr]
                and node.args[1] == torch.bfloat16
            )

        if (
            is_woq_int8_pattern(node)
            or (
                node.target == torch.ops.aten.permute.default
                and len(node.users) == 1
                and is_woq_int8_pattern(next(iter(node.users)))
            )
        ) and _is_const_source(node.args[0]):
            # Case 1: int8_weight -> dq -> bf16_weight
            # Case 2: int8_weight -> permute -> dq -> bf16_weight
            return True

        quant_registered = getattr(torch.ops.quantized_decomposed, "dequantize_per_channel", None) is not None
        if quant_registered and node.target in [
            torch.ops.quantized_decomposed.dequantize_per_channel.default,
            torch.ops.quantized_decomposed.dequantize_per_tensor.default,
            torch.ops.quantized_decomposed.dequantize_per_tensor.tensor,
        ]:
            # For the pattern fp32_weight -> q -> dq
            # We only folding fp32_weight -> q
            # int8_weight and leave dq in graph to be fused
            return True
        return False

    def node_to_last_non_output_use(self) -> Dict[torch.fx.Node, List[torch.fx.Node]]:
        last_non_output_use = collections.defaultdict(list)
        seen_uses = set()
        output_node = next(iter(reversed(self.module.graph.nodes)))

        for node in reversed(self.module.graph.nodes):
            if node.target == "output":
                continue

            def add_use(inp: torch.fx.Node) -> None:
                if inp in seen_uses:
                    return

                seen_uses.add(inp)
                last_non_output_use[node].append(inp)

            # In-place is fine since we don't mutate
            pytree.tree_map_only_(torch.fx.Node, add_use, (node.args, node.kwargs))

            # if this node is only used in output, we want to gc it right away
            if len(node.users) == 1 and output_node in node.users:
                last_non_output_use[node].append(node)

        return last_non_output_use

    def run_node(self, node: torch.fx.Node) -> Any:
        if node.target == "output":
            # because we remove nodes from env on last non output use,
            # re-define them now or we'll get error in interpreter
            def set_env(arg: torch.fx.Node) -> None:
                self.env[arg] = self.unknown_value

            # In-place is fine since we don't mutate
            pytree.tree_map_only_(torch.fx.Node, set_env, node.args)
            return super().run_node(node)

        args, kwargs = self.fetch_args_kwargs_from_env(node)
        flattened_inputs = pytree.arg_tree_leaves(*args, **kwargs)

        # We need to do this weird thing because in cases where flattened_inputs
        # contains a ScriptObject, equality checking results in a type error if
        # the types are different.
        if any(
            type(self.unknown_value) is type(input_) and self.unknown_value == input_ for input_ in flattened_inputs
        ):
            return self.unknown_value

        if node.op == "call_function" and node.target == aten._efficientzerotensor.default:
            return self.unknown_value

        if node.op == "call_function" and node.name == "triton_kernel_wrapper_functional_proxy":
            return self.unknown_value

        # skip constructors, since inductor generates optimal code for them already
        # and turning into tensor would result in an additional global memory read
        if not _is_const_source(node) and not any(isinstance(e, torch.Tensor) for e in flattened_inputs):
            return self.unknown_value

        # All mutations should either be removed or on inputs which we did not make constant
        if isinstance(node.target, torch._ops.OpOverload) and torch.Tag.nondeterministic_seeded in node.target.tags:
            return self.unknown_value

        out = self._deduce_value(node)
        if out == self.unknown_value:
            return self.unknown_value

        if not _is_const_source(node) and isinstance(out, torch.Tensor):
            if out.device.type == "meta":
                return out

            if self.is_impure(node):
                return self.unknown_value

            self.add_node_replacement(node, out)

            flattened_node_inps = pytree.arg_tree_leaves(*node.args, **node.kwargs)

            for n in flattened_node_inps:
                if not isinstance(n, torch.fx.Node):
                    continue

                self.replaced_uses[n] += 1

            for to_delete in self.user_to_last_uses.get(node, []):
                if self.replaced_uses[to_delete] == len(to_delete.users):
                    self.node_replacements.pop(to_delete, None)

        return out

    def add_node_replacement(self, node: torch.fx.Node, tensor: torch.Tensor) -> None:
        self.node_replacements[node] = tensor

    def run(self) -> Any:  # type: ignore[override]
        env: Dict[torch.fx.Node, Any] = {}
        self.insert_placerholder_values(env)
        return super().run(initial_env=env)

    def insert_placerholder_values(self, env: Dict[torch.fx.Node, Any]) -> None:
        for n in self.module.graph.find_nodes(op="placeholder"):
            env[n] = self.unknown_value  # type: ignore[assignment]


def _is_impure(node: torch.fx.Node) -> bool:
    """
    Returns True if the node call affects the model outputs even in case
    the node have zero users, False otherwise.

    :param node: A node to check.
    :return: True if the node call affects the model outputs even in case
        the node have zero users, False otherwise.

    """
    return node.op in {"placeholder", "output"}


def constant_fold(
    gm: torch.fx.GraphModule,
    constraint_fn: Optional[Callable[[torch.fx.Node], bool]] = None,
) -> None:
    """
    Calcualtes constant subgraphs values and replaces them with a constant node inplace.

    :param gm: Given graph model.
    :param constraint_fn: Constraint function which takes a node and returs the constraint:
        should the node be constant folded or not.
    """
    with torch.no_grad():
        with torch.utils._python_dispatch._disable_current_modes():
            cf = ConstantFolder(gm)
            cf.run()

            for node, constant in cf.node_replacements.items():
                if constraint_fn is not None and not constraint_fn(node):
                    continue
                _replace_node_with_constant(gm, node, constant)

            erased_params = []
            for node in gm.graph.find_nodes(op="get_attr"):
                if len(node.users) == 0:
                    if hasattr(gm, node.target):
                        delattr(gm, node.target)
                    erased_params.append(node)

            for node in erased_params:
                gm.graph.erase_node(node)

            # Custom _is_impure function allows to eliminate all layers with zero
            # users including inplace ops like relu_ besides output and placeholders.
            gm.graph.eliminate_dead_code(_is_impure)
            gm.graph.lint()
            gm.recompile()



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from abc import abstractmethod
from typing import Dict, Union

import nncf
from nncf.api.compression import CompressionAlgorithmController
from nncf.common.accuracy_aware_training.runner import BaseAccuracyAwareTrainingRunner
from nncf.common.accuracy_aware_training.runner import BaseAdaptiveCompressionLevelTrainingRunner
from nncf.common.accuracy_aware_training.runner import TrainingRunner
from nncf.common.utils.backend import BackendType
from nncf.common.utils.backend import get_backend


class TrainingRunnerCreator(ABC):
    """
    Declares the factory method returning TrainingRunner object
    """

    @abstractmethod
    def create_training_loop(self) -> TrainingRunner:
        pass


class EarlyExitTrainingRunnerCreator(TrainingRunnerCreator):
    """
    Class creates an Early Exit Training Runner depending on an used backend.
    """

    def __init__(
        self,
        accuracy_aware_training_params: Dict[str, Union[float, int]],
        compression_controller: CompressionAlgorithmController,
        uncompressed_model_accuracy: float,
        verbose: bool,
        dump_checkpoints: bool,
        lr_updates_needed: bool,
    ):
        self.accuracy_aware_training_params = accuracy_aware_training_params
        self.compression_controller = compression_controller
        self.lr_updates_needed = lr_updates_needed
        self.verbose = verbose
        self.dump_checkpoints = dump_checkpoints
        self.uncompressed_model_accuracy = uncompressed_model_accuracy

    def create_training_loop(self) -> BaseAccuracyAwareTrainingRunner:
        """
        Creates an object of AccuracyAwareTrainingRunner depending on user backend

        :return: AccuracyAwareTrainingRunner object
        """
        nncf_backend = get_backend(self.compression_controller.model)  # type: ignore
        if nncf_backend is BackendType.TORCH:
            from nncf.torch.accuracy_aware_training.runner import PTAccuracyAwareTrainingRunner

            return PTAccuracyAwareTrainingRunner(
                self.accuracy_aware_training_params,
                self.uncompressed_model_accuracy,
                self.verbose,
                self.dump_checkpoints,
                self.lr_updates_needed,
            )
        if nncf_backend == BackendType.TENSORFLOW:
            from nncf.tensorflow.accuracy_aware_training.runner import TFAccuracyAwareTrainingRunner

            return TFAccuracyAwareTrainingRunner(
                self.accuracy_aware_training_params,
                self.uncompressed_model_accuracy,
                self.verbose,
                self.dump_checkpoints,
                self.lr_updates_needed,
            )
        raise nncf.UnsupportedBackendError("Got an unsupported value of nncf_backend")


class AdaptiveCompressionLevelTrainingRunnerCreator(TrainingRunnerCreator):
    """
    Class creates an Adaptive Compression Level Training Runner depending on an used backend.
    """

    def __init__(
        self,
        accuracy_aware_training_params: Dict[str, Union[float, int]],
        compression_controller: CompressionAlgorithmController,
        uncompressed_model_accuracy: float,
        verbose: bool,
        dump_checkpoints: bool,
        lr_updates_needed: bool,
        minimal_compression_rate: float,
        maximal_compression_rate: float,
    ):
        self.accuracy_aware_training_params = accuracy_aware_training_params
        self.compression_controller = compression_controller
        self.uncompressed_model_accuracy = uncompressed_model_accuracy
        self.lr_updates_needed = lr_updates_needed
        self.verbose = verbose
        self.minimal_compression_rate = minimal_compression_rate
        self.maximal_compression_rate = maximal_compression_rate
        self.dump_checkpoints = dump_checkpoints

    def create_training_loop(self) -> BaseAdaptiveCompressionLevelTrainingRunner:
        """
        Creates an object of AdaptiveCompressionLevelTrainingRunner depending on user backend

        :return: AdaptiveCompressionLevelTrainingRunner object
        """
        nncf_backend = get_backend(self.compression_controller.model)  # type: ignore

        if nncf_backend is BackendType.TORCH:
            from nncf.torch.accuracy_aware_training.runner import PTAdaptiveCompressionLevelTrainingRunner

            return PTAdaptiveCompressionLevelTrainingRunner(
                self.accuracy_aware_training_params,
                self.uncompressed_model_accuracy,
                self.verbose,
                self.dump_checkpoints,
                self.lr_updates_needed,
                self.minimal_compression_rate,
                self.maximal_compression_rate,
            )
        if nncf_backend == BackendType.TENSORFLOW:
            from nncf.tensorflow.accuracy_aware_training.runner import TFAdaptiveCompressionLevelTrainingRunner

            return TFAdaptiveCompressionLevelTrainingRunner(
                self.accuracy_aware_training_params,
                self.uncompressed_model_accuracy,
                self.verbose,
                self.dump_checkpoints,
                self.lr_updates_needed,
                self.minimal_compression_rate,
                self.maximal_compression_rate,
            )
        raise nncf.UnsupportedBackendError("Got an unsupported value of nncf_backend")



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import random
from typing import Any, Callable, Dict, List, Optional

import torch
import torch.nn.functional as F
from torch import nn
from torch.nn import Parameter

from nncf.common.graph import NNCFNode
from nncf.common.graph import NNCFNodeName
from nncf.common.graph.layer_attributes import ConvolutionLayerAttributes
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.graph.transformations.commands import TransformationCommand
from nncf.common.graph.transformations.commands import TransformationPriority
from nncf.common.logging import nncf_logger
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.base_handler import ELASTICITY_BUILDERS
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.base_handler import ELASTICITY_HANDLERS_MAP
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.base_handler import ELASTICITY_PARAMS
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.base_handler import BaseElasticityParams
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.base_handler import SingleElasticityBuilder
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.base_handler import SingleElasticityHandler
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.elasticity_dim import ElasticityDim
from nncf.torch.graph.transformations.commands import PTInsertionCommand
from nncf.torch.graph.transformations.commands import PTTargetPoint
from nncf.torch.layers import NNCFConv2d
from nncf.torch.module_operations import UpdateInputs
from nncf.torch.module_operations import UpdatePadding
from nncf.torch.module_operations import UpdateWeight
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.utils import is_tracing_state
from nncf.torch.utils import no_jit_trace

KernelSizeType = int
ElasticKernelConfig = List[KernelSizeType]  # list of kernel sizes per layer
ElasticKernelSearchSpace = List[List[KernelSizeType]]


class EKParamsStateNames:
    MAX_NUM_KERNELS = "max_num_kernels"


@ELASTICITY_PARAMS.register(ElasticityDim.KERNEL)
class ElasticKernelParams(BaseElasticityParams):
    _state_names = EKParamsStateNames

    def __init__(self, max_num_kernels: int = -1):
        """
        Constructor

        :param max_num_kernels: Restricts total number of different elastic kernel values for each layer.
        The default value is -1 means that there's no restrictions.
        """
        self.max_num_kernels = max_num_kernels

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "ElasticKernelParams":
        """
        Creates the object from its config.
        """
        kwargs = {
            cls._state_names.MAX_NUM_KERNELS: config.get(cls._state_names.MAX_NUM_KERNELS, -1),
        }
        return cls(**kwargs)

    @classmethod
    def from_state(cls, state: Dict[str, Any]) -> "ElasticKernelParams":
        """
        Creates the object from its state.

        :param state: Output of `get_state()` method.
        """
        return cls(**state)

    def get_state(self) -> Dict[str, Any]:
        """
        Returns the compression loss state.

        :return: The compression loss state.
        """
        return {
            self._state_names.MAX_NUM_KERNELS: self.max_num_kernels,
        }

    def __eq__(self, other: "ElasticKernelParams") -> bool:
        return self.__dict__ == other.__dict__


class ElasticKernelOp:
    """
    Base class for introducing elastic kernel for the operations. On the forward pass it takes parameters of operations
    and modifies in the way that kernel size is changing to a given value.
    """

    def __init__(self, max_kernel_size: KernelSizeType, node_name: NNCFNodeName):
        """
        Constructor.

        :param max_kernel_size: maximum kernel size value in the original operation.
        :param node_name: string representation of operation address. It's used for more informative messages only.
        """
        super().__init__()
        self._kernel_size_list = []
        self._active_kernel_size = max_kernel_size
        self._max_kernel_size = max_kernel_size
        self._node_name = node_name

    def get_active_kernel_size(self) -> KernelSizeType:
        """
        :return: a target kernel size that operation's parameters should have after forward's call
        """
        return self._active_kernel_size

    def set_active_kernel_size(self, kernel_size: KernelSizeType) -> None:
        """
        Sets current level of elasticity for the operation - kernel size value that parameters should have.
        The actual modification of parameters happens on forward call.
        The value should be less the original kernel size and more than one.

        :param kernel_size: kernel size value
        """
        if kernel_size is None or kernel_size > self.max_kernel_size or kernel_size < 1:
            raise AttributeError(
                "Invalid kernel size={} in scope={}.\nIt should be within the range: [1, {}]".format(
                    kernel_size, self.node_name, self.max_kernel_size
                )
            )

        self._active_kernel_size = kernel_size

    @property
    def kernel_size_list(self) -> List[KernelSizeType]:
        """
        Gets list of all available kernel sizes to select from. Each value corresponds to a single element in the
        search space of operation. The search space of the model is cartesian product of search spaces of operation.

        :return: list of kernel sizes
        """
        return self._kernel_size_list

    @property
    def node_name(self) -> NNCFNodeName:
        """
        :return: node name that corresponds to operation with elastic kernel
        """
        return self._node_name

    @property
    def max_kernel_size(self) -> KernelSizeType:
        """
        :return: kernel size in the original operation
        """
        return self._max_kernel_size


class ElasticKernelConv2DOp(ElasticKernelOp, nn.Module):
    """
    Introduces elastic kernel for the 2D convolution. On the forward pass it takes parameters of operations
    and modifies in the way that kernel size is changing to a given value.
    """

    def __init__(
        self,
        max_kernel_size: KernelSizeType,
        node_name: NNCFNodeName,
        params: ElasticKernelParams,
        original_padding_value: Optional[int] = 0,
    ):
        """
        Constructor.

        :param max_kernel_size: maximum kernel size value in the original operation.
        :param node_name: string representation of operation address. It's used for more informative messages only.
        :param params: parameters to configure elastic kernel for the operation.
        :param original_padding_value: the padding value used in the original model.
        """
        super().__init__(max_kernel_size=max_kernel_size, node_name=node_name)
        self._max_num_params = params.max_num_kernels
        self._original_padding_value = original_padding_value
        # Create kernel_size_list based on max module kernel size
        self._kernel_size_list = self.generate_kernel_size_list(max_kernel_size, original_padding_value)
        self._ks_set = list(set(self.kernel_size_list))
        self._ks_set.sort()

        scale_params = {}
        for i in range(len(self._ks_set) - 1):
            ks_small = self._ks_set[i]
            ks_larger = self._ks_set[i + 1]
            param_name = "%dto%d" % (ks_larger, ks_small)
            # noinspection PyArgumentList
            scale_params["%s_matrix" % param_name] = Parameter(torch.eye(ks_small**2))
        for name, param in scale_params.items():
            self.register_parameter(name, param)

    def generate_kernel_size_list(
        self, max_kernel_size: KernelSizeType, original_padding_value: int
    ) -> List[KernelSizeType]:
        """
        Generates list of available kernel size values.

        :param max_kernel_size: maximum value of kernel size, it's supposed to be odd
        :param original_padding_value: the padding value used in the original model.
        :return: list of kernel size values.
        """
        DEFAULT_KERNEL_SIZE_STEP = 2
        assert max_kernel_size % 2 > 0, "kernel size should be odd number"
        if max_kernel_size == 1:
            return [1]
        kernel = max_kernel_size
        ks_list = []
        while kernel >= max(max_kernel_size - 2 * original_padding_value, 3):
            ks_list.append(kernel)
            kernel -= DEFAULT_KERNEL_SIZE_STEP
            if self._max_num_params == len(ks_list):
                break
        return ks_list

    def forward(self, weight: torch.Tensor) -> torch.Tensor:
        """
        Modifies weight to have kernel size equals to active kernel size value.

        :param weight: weight tensor to be modified
        :return: modified weight
        """
        kernel_size = self.get_active_kernel_size()
        nncf_logger.debug(f"Conv2d with active kernel size={kernel_size} in scope={self.node_name}")

        result = weight
        if is_tracing_state():
            with no_jit_trace():
                if kernel_size > 1:
                    result = self._get_active_filter(kernel_size, weight).detach()
        else:
            if kernel_size > 1:
                result = self._get_active_filter(kernel_size, weight)
        return result

    def set_active_kernel_size(self, kernel_size: KernelSizeType) -> None:
        """
        Sets current level of elasticity for the operation - kernel size value that parameters should have.
        The actual modification of parameters happens on forward call.
        The value should be less the original kernel size and more than one.

        :param kernel_size: kernel size value
        """
        nncf_logger.debug(f"set active elastic_kernel={kernel_size} in scope={self.node_name}")
        assert kernel_size % 2 > 0, "kernel size should be odd number"
        if kernel_size not in self.kernel_size_list and kernel_size != self.max_kernel_size:
            raise ValueError("invalid kernel size to set. Should be a number in {}".format(self.kernel_size_list))
        super().set_active_kernel_size(kernel_size)

    @staticmethod
    def _sub_filter_start_end(kernel_size, sub_kernel_size):
        center = kernel_size // 2
        dev = sub_kernel_size // 2
        start, end = center - dev, center + dev + 1
        assert end - start == sub_kernel_size
        return start, end

    def _get_active_filter(self, kernel_size, weight):
        max_kernel_size = max(self.kernel_size_list)
        start, end = self._sub_filter_start_end(max_kernel_size, kernel_size)
        filters = weight[:, :, start:end, start:end]
        if kernel_size < max_kernel_size:
            start_filter = weight  # start with max kernel
            for i in range(len(self._ks_set) - 1, 0, -1):
                src_ks = self._ks_set[i]
                if src_ks <= kernel_size:
                    break
                target_ks = self._ks_set[i - 1]
                start, end = self._sub_filter_start_end(src_ks, target_ks)
                _input_filter = start_filter[:, :, start:end, start:end]
                _input_filter = _input_filter.contiguous()
                _input_filter = _input_filter.view(_input_filter.size(0), _input_filter.size(1), -1)
                _input_filter = _input_filter.view(-1, _input_filter.size(2))
                _input_filter = F.linear(
                    _input_filter,
                    self.__getattr__("%dto%d_matrix" % (src_ks, target_ks)),
                )
                _input_filter = _input_filter.view(filters.size(0), filters.size(1), target_ks**2)
                _input_filter = _input_filter.view(filters.size(0), filters.size(1), target_ks, target_ks)
                start_filter = _input_filter
            filters = start_filter
        return filters

    @property
    def original_padding_value(self) -> int:
        return self._original_padding_value


class ElasticKernelPaddingAdjustment:
    """
    Auxiliary operation that adjusts padding of the operation with elastic kernel.
    This adjustment ensures the same output shapes for different kernel sizes and thus frees from the need to adapt the
    rest of the operation in the model for each new value of kernel size.
    """

    def __init__(self, elastic_k_w_op: ElasticKernelConv2DOp):
        self._elastic_k_w_op = elastic_k_w_op

    def __call__(self, _) -> int:
        shift_padding_value = (
            self._elastic_k_w_op.max_kernel_size - self._elastic_k_w_op.get_active_kernel_size()
        ) // 2
        return self._elastic_k_w_op.original_padding_value - shift_padding_value


class ElasticKernelInputForExternalPadding:
    """
    Auxiliary operation that adjusts inputs for operations that implements same padding type
    (e.g. NNCFUserConv2dStaticSamePadding).
    This adjustment ensures the same output shapes for different kernel sizes and thus frees from the need to adapt the
    rest of the operation in the model for each new value of kernel size.
    """

    def __init__(self, elastic_k_w_op: ElasticKernelConv2DOp, max_kernel_size: KernelSizeType):
        self._elastic_k_w_op = elastic_k_w_op
        self._max_kernel_size = max_kernel_size

    def __call__(self, x) -> int:
        active_kernel_size = self._elastic_k_w_op.get_active_kernel_size()
        diff = (self._max_kernel_size - active_kernel_size) // 2
        h = x.size(2)
        w = x.size(3)
        new_x = x[:, :, diff : h - diff, diff : w - diff]
        return new_x


class ElasticKernelHandler(SingleElasticityHandler):
    """
    An interface for handling elastic kernel dimension in the network, i.e. define size of kernels in the conv layers.
    """

    def __init__(self, elastic_kernel_ops: List[ElasticKernelOp], transformation_commands: List[TransformationCommand]):
        super().__init__()
        self._elastic_kernel_ops = elastic_kernel_ops
        self._transformation_commands = transformation_commands

    def get_transformation_commands(self) -> List[TransformationCommand]:
        """
        :return: transformation commands for introducing the elasticity to NNCFNetwork
        """
        return self._transformation_commands

    def get_search_space(self) -> ElasticKernelSearchSpace:
        """
        :return: search space that is produced by iterating over all elastic kernel parameters
        """
        return self._collect_ops_data_by_selection_rule(lambda op: op.kernel_size_list)

    def get_active_config(self) -> ElasticKernelConfig:
        """
        Forms an elasticity configuration that describes currently activated Subnet

        :return: list of kernel sizes per layer
        """
        return self._collect_ops_data_by_selection_rule(lambda op: op.get_active_kernel_size())

    def get_random_config(self) -> ElasticKernelConfig:
        """
        Forms an elasticity configuration that describes a Subnet with randomly chosen elastic kernels

        :return: list of kernel sizes per layer
        """
        config = self._collect_ops_data_by_selection_rule(
            lambda op: op.kernel_size_list[random.randrange(0, len(op.kernel_size_list))]  # nosec
        )
        return config

    def get_minimum_config(self) -> ElasticKernelConfig:
        """
        Forms an elasticity configuration that describes a Subnet with minimum elastic kernels

        :return: list of kernel sizes per layer
        """
        return self._collect_ops_data_by_selection_rule(lambda op: min(op.kernel_size_list))

    def get_maximum_config(self) -> ElasticKernelConfig:
        """
        Forms an elasticity configuration that describes a Subnet with maximum elastic kernels

        :return: list of kernel sizes per layer
        """
        return self._collect_ops_data_by_selection_rule(lambda op: max(op.kernel_size_list))

    def activate_supernet(self) -> None:
        """
        Activates the Supernet - the original network to which elasticity was applied.
        """
        supernet_config = self._collect_ops_data_by_selection_rule(lambda op: op.max_kernel_size)
        self.activate_subnet_for_config(supernet_config)

    def activate_subnet_for_config(self, config: ElasticKernelConfig) -> None:
        """
        Activates a Subnet that corresponds to the given elasticity configuration

        :return: list of kernel sizes per layer
        """
        for op, ks in zip(self._elastic_kernel_ops, config):
            op.set_active_kernel_size(ks)

    def get_active_kernel_sizes_per_node(self) -> Dict[str, Any]:
        """
        :return: mapping of node name to the active kernel sizes in that node
        """
        active_kernel_sizes = {}
        for elastic_kernel_op in self._elastic_kernel_ops:
            active_kernel_size = elastic_kernel_op.get_active_kernel_size()
            active_kernel_sizes[elastic_kernel_op.node_name] = (active_kernel_size, active_kernel_size)
        return active_kernel_sizes

    def resolve_conflicts_with_other_elasticities(
        self, config: ElasticKernelConfig, elasticity_handlers: ELASTICITY_HANDLERS_MAP
    ) -> ElasticKernelConfig:
        """
        Resolves a conflict between the given elasticity config and active elasticity configs of the given handlers.
        For example, elastic width configuration may contradict to elastic depth one. When we activate some
        configuration in the Elastic Width Handler, i.e. define number of output channels for some layers, we
        change output shapes of the layers. Consequently, it affects the blocks that can be skipped by Elastic Depth
        Handler, because input and output shapes may not be identical now.

        :param config: elasticity configuration
        :param elasticity_handlers: map of elasticity dimension to elasticity handler
        :return: elasticity configuration without conflicts with other active configs of other elasticity handlers
        """
        return config

    def _collect_ops_data_by_selection_rule(self, selection_rule: Callable) -> List[Any]:
        return list(map(selection_rule, self._elastic_kernel_ops))


class EKBuilderStateNames:
    NODE_NAMES_TO_MAKE_ELASTIC = "node_names_to_make_elastic"


@ELASTICITY_BUILDERS.register(ElasticityDim.KERNEL)
class ElasticKernelBuilder(SingleElasticityBuilder):
    _state_names = EKBuilderStateNames

    def __init__(
        self,
        params: ElasticKernelParams,
        ignored_scopes: Optional[List[str]] = None,
        target_scopes: Optional[List[str]] = None,
    ):
        super().__init__(ignored_scopes, target_scopes)
        self._node_names_to_make_elastic: List[NNCFNodeName] = []
        self._params = params

    def build(self, target_model: NNCFNetwork) -> ElasticKernelHandler:
        """
        Creates modifications to the given NNCFNetwork for introducing elastic kernel and creates a handler object that
        can manipulate this elasticity.

        :param target_model: a target NNCFNetwork for adding modifications
        :return: a handler object that can manipulate the elastic kernel.
        """
        elastic_kernel_ops: List[ElasticKernelOp] = []
        transformation_commands = []

        graph = target_model.nncf.get_original_graph()
        device = next(target_model.parameters()).device
        pad_commands = []

        if not self._node_names_to_make_elastic:
            elastic_kernel_types = [NNCFConv2d.op_func_name]
            all_elastic_kernel_nodes: List[NNCFNode] = graph.get_nodes_by_types(elastic_kernel_types)
            self._node_names_to_make_elastic = [node.node_name for node in all_elastic_kernel_nodes]

        for node_name in self._node_names_to_make_elastic:
            nncf_logger.debug(f"Adding Elastic Kernel op for Conv2D in scope: {node_name}")
            node = graph.get_node_by_name(node_name)
            layer_attrs = node.layer_attributes
            assert isinstance(layer_attrs, ConvolutionLayerAttributes), "Conv2D can have elastic kernel only"
            max_kernel_size = layer_attrs.kernel_size[0]
            original_padding_values = layer_attrs.padding_values[0]
            elastic_kernel_op = ElasticKernelConv2DOp(max_kernel_size, node_name, self._params, original_padding_values)
            elastic_kernel_op.to(device)
            update_conv_params_op = UpdateWeight(elastic_kernel_op)
            transformation_commands.append(
                PTInsertionCommand(
                    PTTargetPoint(TargetType.PRE_LAYER_OPERATION, target_node_name=node_name),
                    update_conv_params_op,
                    TransformationPriority.PRUNING_PRIORITY,
                )
            )
            # TODO(nlyalyus): ticket 71613: remove hardcode for EfficientNet
            #   should be pairing of pad and conv operations by graph analysis
            #   get
            #       1) input size before pad
            #       2) input size after pad
            #       3) active kernel size
            #       4) max kernel size
            #       5) built-in conv padding
            #   output
            #       how to change padded input (e.g. center crop)
            if "NNCFUserConv2dStaticSamePadding" in node.node_name:
                if max_kernel_size >= 3:
                    crop_op = ElasticKernelInputForExternalPadding(elastic_kernel_op, max_kernel_size)
                    op = UpdateInputs(crop_op).to(device)
                    nncf_logger.debug(f"Padded input will be cropped for {node_name}")
                    pad_commands.append(
                        PTInsertionCommand(
                            PTTargetPoint(target_type=TargetType.PRE_LAYER_OPERATION, target_node_name=node_name),
                            op,
                            TransformationPriority.DEFAULT_PRIORITY,
                        )
                    )
            else:
                # Padding
                ap = ElasticKernelPaddingAdjustment(elastic_kernel_op)
                pad_op = UpdatePadding(ap).to(device)
                nncf_logger.debug(f"Padding will be adjusted for {node_name}")
                pad_commands.append(
                    PTInsertionCommand(
                        PTTargetPoint(target_type=TargetType.PRE_LAYER_OPERATION, target_node_name=node_name),
                        pad_op,
                        TransformationPriority.DEFAULT_PRIORITY,
                    )
                )
            elastic_kernel_ops.append(elastic_kernel_op)
        if pad_commands:
            transformation_commands += pad_commands

        return ElasticKernelHandler(elastic_kernel_ops, transformation_commands)

    def load_state(self, state: Dict[str, Any]) -> None:
        """
        Initializes object from the state.

        :param state: Output of `get_state()` method.
        """
        params_from_state = state[SingleElasticityBuilder._state_names.ELASTICITY_PARAMS]
        params = ElasticKernelParams.from_state(params_from_state)
        if self._params and self._params != params:
            nncf_logger.warning(
                "Different elasticity parameters were provided in two places: on init and on loading "
                "state. The one from state is taken by ignoring the ones from init."
            )
        self._params = params
        self._node_names_to_make_elastic = state[self._state_names.NODE_NAMES_TO_MAKE_ELASTIC]

    def get_state(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.

        :return: state of the object
        """
        return {
            SingleElasticityBuilder._state_names.ELASTICITY_PARAMS: self._params.get_state(),
            self._state_names.NODE_NAMES_TO_MAKE_ELASTIC: self._node_names_to_make_elastic,
        }



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple, Union

import tensorflow as tf

from nncf.common.tensor_statistics.collectors import ReductionAxes
from nncf.tensorflow.layers.data_layout import get_weight_shape


def get_axes(ndims: int, per_channel: bool, channel_axes: Union[int, list, tuple]) -> list:
    axes = list(range(ndims))
    if per_channel:
        for val in channel_axes:
            val = (ndims + val) % ndims
            axes.remove(val)
    return axes


def get_reduction_shape_activations(
    layer: tf.keras.layers.Layer, channel_axes: Union[int, tuple, list], use_per_sample_stats: bool
) -> ReductionAxes:
    ndims = len(layer.get_input_shape_at(0))
    channel_axes_ = channel_axes if isinstance(channel_axes, (list, tuple)) else [channel_axes]
    reduction_shape = get_axes(ndims, layer.per_channel, channel_axes_)
    if use_per_sample_stats:
        reduction_shape = reduction_shape[1:]
    return tuple(reduction_shape)


def get_reduction_shape_weights(
    layer: tf.keras.layers.Layer, weight_attr: str, channel_axes: Union[int, tuple, list], per_channel: bool
) -> ReductionAxes:
    weight_shape = get_weight_shape(layer, weight_attr)
    ndims = len(weight_shape)
    channel_axes_ = channel_axes if isinstance(channel_axes, (list, tuple)) else [channel_axes]
    reduction_shape = get_axes(ndims, per_channel, channel_axes_)
    return tuple(reduction_shape)


def convert_rs_to_pt_type(input_shape: Tuple[int], reduction_shape: ReductionAxes) -> ReductionAxes:
    if len(reduction_shape) == len(input_shape):
        pt_reduction_shape = [1]
    else:
        pt_reduction_shape = []
        for dim_idx, dim in enumerate(input_shape):
            if dim_idx in reduction_shape:
                pt_reduction_shape.append(1)
            else:
                pt_reduction_shape.append(dim)
    return tuple(pt_reduction_shape)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from itertools import islice

import tensorflow as tf

from nncf.common.initialization.batchnorm_adaptation import BatchnormAdaptationAlgorithmImpl
from nncf.common.logging.progress_bar import ProgressBar
from nncf.tensorflow.graph.metatypes.keras_layers import TFBatchNormalizationLayerMetatype
from nncf.tensorflow.graph.metatypes.matcher import get_keras_layer_metatype


class BNTrainingStateSwitcher:
    """
    Context manager for switching BatchNormalization layer train mode.
    At the enter, set BatchNormalization layers to the train mode.
    At the exit, restore original BatchNormalization layer mode.
    """

    def __init__(self, model: tf.keras.Model):
        self._model = model
        self._original_training_state = {}

    def __enter__(self):
        for layer in self._model.layers:
            if get_keras_layer_metatype(layer) == TFBatchNormalizationLayerMetatype:
                self._original_training_state[layer] = layer.trainable
                layer.trainable = True

    def __exit__(self, *exc):
        for layer in self._model.layers:
            if get_keras_layer_metatype(layer) == TFBatchNormalizationLayerMetatype:
                layer.trainable = self._original_training_state[layer]


class TFBatchnormAdaptationAlgorithmImpl(BatchnormAdaptationAlgorithmImpl):
    """
    Implementation of the batch-norm statistics adaptation algorithm for the TensorFlow backend.
    """

    def run(self, model: tf.keras.Model) -> None:
        """
        Runs the batch-norm statistics adaptation algorithm.

        :param model: A model for which the algorithm will be applied.
        """
        if self._device is not None:
            raise ValueError(
                "TF implementation of batchnorm adaptation algorithm "
                "does not support switch of devices. Model initial device "
                "is used by default for batchnorm adaptation."
            )
        with BNTrainingStateSwitcher(model):
            for x, _ in ProgressBar(
                islice(self._data_loader, self._num_bn_adaptation_steps),
                total=self._num_bn_adaptation_steps,
                desc="BatchNorm statistics adaptation",
            ):
                model(x, training=True)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional, Tuple

import nncf
from nncf import NNCFConfig
from nncf.api.compression import CompressionAlgorithmBuilder
from nncf.api.compression import CompressionAlgorithmController
from nncf.api.compression import CompressionLoss
from nncf.api.compression import CompressionScheduler
from nncf.api.compression import CompressionStage
from nncf.api.compression import TModel
from nncf.common.statistics import NNCFStatistics
from nncf.common.utils.backend import BackendType
from nncf.common.utils.backend import copy_model
from nncf.common.utils.backend import get_backend


class CompositeCompressionLoss(CompressionLoss):
    """
    The `CompositeCompressionLoss` class stores a group of `CompressionLoss`
    instances as a list of children that are treated the same way as a single
    `CompressionLoss` instance.
    """

    def __init__(self):
        super().__init__()
        self._child_losses = []

    @property
    def child_losses(self) -> List[CompressionLoss]:
        return self._child_losses

    def add(self, child_loss: CompressionLoss) -> None:
        """
        Add `CompressionLoss` instance to the list of children.

        :param child_loss: A `CompressionLoss` instance.
        """
        self._child_losses.append(child_loss)

    def load_state(self, state: List[Dict[str, Any]]) -> None:
        """
        Loads the composite compression loss state.

        :param state: Output of `get_state()` method.
        """
        for child_loss, child_state in zip(self._child_losses, state):
            child_loss.load_state(child_state)

    def get_state(self) -> List[Dict[str, Any]]:
        """
        Returns the composite compression loss state.

        :return: The composite compression loss state.
        """
        composite_state = []
        for child_loss in self.child_losses:
            composite_state.append(child_loss.get_state())
        return composite_state

    def calculate(self, *args, **kwargs) -> Any:
        """
        Traverses through all children and calculates the total compression
        loss value.

        :return: The compression loss value.
        """

        if len(self._child_losses) == 0:
            raise nncf.InternalError("Cannot calculate the loss value because the number of child loss is 0.")

        result_loss = 0
        for loss in self._child_losses:
            result_loss += loss()
        return result_loss


class CompositeCompressionScheduler(CompressionScheduler):
    """
    The `CompositeCompressionScheduler` class stores a group of `CompressionScheduler`
    instances as a list of children that are treated the same way as a single
    `CompressionScheduler` instance.
    """

    def __init__(self):
        super().__init__()
        self._child_schedulers = []

    @property
    def child_schedulers(self) -> List[CompressionScheduler]:
        return self._child_schedulers

    def add(self, child_scheduler: CompressionScheduler) -> None:
        """
        Add `CompressionScheduler` instance to the list of children.

        :param child_scheduler: A `CompressionScheduler` instance.
        """
        self._child_schedulers.append(child_scheduler)

    def step(self, next_step: Optional[int] = None) -> None:
        """
        Calls step() method for all children.

        :param next_step: The global step index for which the compression scheduler
            will update the state of the compression method.
        """
        super().step(next_step)
        for scheduler in self._child_schedulers:
            scheduler.step(next_step)

    def epoch_step(self, next_epoch: Optional[int] = None) -> None:
        """
        Calls epoch_step() method for all children.

        :param next_epoch: The epoch index for which the compression scheduler
            will update the state of the compression method.
        """
        super().epoch_step(next_epoch)
        for scheduler in self._child_schedulers:
            scheduler.epoch_step(next_epoch)

    def load_state(self, state: List[Dict[str, Any]]) -> None:
        """
        Calls `load_state()` method for all children.

        :param state: Output of `get_state()` method.
        """
        for child_scheduler, child_state in zip(self._child_schedulers, state):
            child_scheduler.load_state(child_state)

    def get_state(self) -> List[Dict[str, Any]]:
        """
        Returns the composite compression scheduler state. This state contains
        the state of all children.

        :return: The composite compression scheduler state.
        """
        composite_state = []
        for child_scheduler in self._child_schedulers:
            composite_state.append(child_scheduler.get_state())
        return composite_state


class CompositeCompressionAlgorithmController(CompressionAlgorithmController):
    """
    The `CompositeCompressionAlgorithmController` class stores a group of
    `CompressionAlgorithmController` instances as a list of children that are
    treated the same way as a single `CompressionAlgorithmController` instance.
    """

    BUILDER_STATE = "builder_state"
    CONTROLLER_STATE = "ctrl_state"

    def __init__(self, target_model: TModel):
        """
        Initializes the internal state of the composite compression algorithm
        controller.

        :param target_model: The model with additional modifications necessary
            to enable algorithm-specific compression during fine-tuning built
            by the `CompressionAlgorithmBuilder`.
        """
        super().__init__(target_model)
        self._child_ctrls = []
        self._loss = CompositeCompressionLoss()
        self._scheduler = CompositeCompressionScheduler()
        self._builder_state = None
        self._name = None

    @property
    def loss(self) -> CompressionLoss:
        return self._loss

    @property
    def scheduler(self) -> CompressionScheduler:
        return self._scheduler

    @property
    def child_ctrls(self) -> List[CompressionAlgorithmController]:
        return self._child_ctrls

    @property
    def name(self) -> str:
        raise self._name

    def add(self, child_ctrl: CompressionAlgorithmController) -> None:
        """
        Add `CompressionAlgorithmController` instance to the list of children.

        :param child_ctrl: A `CompressionAlgorithmController` instance.
        """
        if child_ctrl.model is not self.model:
            raise nncf.InternalError(
                "Cannot create a composite controller from controllers belonging to different models!"
            )

        self._child_ctrls.append(child_ctrl)
        self._loss.add(child_ctrl.loss)
        self._scheduler.add(child_ctrl.scheduler)

    def compression_stage(self) -> CompressionStage:
        """
        Returns the compression stage. Should be used on saving best checkpoints
        to distinguish between uncompressed, partially compressed, and fully
        compressed models.

        :return: The compression stage of the target model.
        """
        if not self.child_ctrls:
            return CompressionStage.UNCOMPRESSED
        result = None
        for ctrl in self.child_ctrls:
            current_level = ctrl.compression_stage()
            if result is None:
                result = current_level
            else:
                result += current_level
        return result

    def load_state(self, state: Dict[str, Dict[str, Any]]) -> None:
        """
        Loads the composite compression controller state from the map of algorithm name to the dictionary with state
        attributes.

        :param state: map of the algorithm name to the dictionary with the corresponding state attributes.
        """
        for ctrl in self.child_ctrls:
            ctrl.load_state(state)

    def get_state(self) -> Dict[str, Dict[str, Any]]:
        """
        Returns composite compression controller state, which is the map of the algorithm name to the dictionary with
        the corresponding state attributes. This state contains the state of all children.

        :return: The composite compression controller state.
        """
        result = {}
        for ctrl in self.child_ctrls:
            result.update(ctrl.get_state())
        return result

    def statistics(self, quickly_collected_only: bool = False) -> NNCFStatistics:
        """
        Returns a `NNCFStatistics` class instance.

        :param quickly_collected_only: Enables collection of the statistics that
            don't take too much time to compute. Can be helpful for the case when
            need to keep track of statistics on each training batch/step/iteration.
        :return: A `NNCFStatistics` class instance.
        """
        nncf_stats = NNCFStatistics()

        for ctrl in self.child_ctrls:
            ctrl_stats = ctrl.statistics(quickly_collected_only)
            for algorithm_name, stats in ctrl_stats:
                nncf_stats.register(algorithm_name, stats)

        return nncf_stats

    def prepare_for_export(self) -> None:
        """
        Prepare the compressed model for deployment.
        """
        stripped_model = self._model
        for ctrl in self.child_ctrls:
            stripped_model = ctrl.strip_model(stripped_model)
        self._model = stripped_model

    def strip(self, do_copy: bool = True) -> TModel:
        model = self.model
        if do_copy:
            model = copy_model(model)
        for ctrl in self.child_ctrls:
            model = ctrl.strip_model(model, do_copy=False)
        return model

    @property
    def compression_rate(self) -> float:
        raise NotImplementedError

    @compression_rate.setter
    def compression_rate(self, compression_rate: float) -> None:
        raise NotImplementedError

    @property
    def maximal_compression_rate(self) -> float:
        return min(child_ctrl.maximal_compression_rate for child_ctrl in self.child_ctrls)

    def export_model(
        self,
        save_path: str,
        save_format: Optional[str] = None,
        input_names: Optional[List[str]] = None,
        output_names: Optional[List[str]] = None,
        model_args: Optional[Tuple[Any, ...]] = None,
    ) -> None:
        """
        Exports the compressed model to the specified format for deployment.

        Makes method-specific preparations of the model, (e.g. removing auxiliary
        layers that were used for the model compression), then exports the model to
        the specified path.

        :param save_path: The path where the model will be saved.
        :param save_format: Saving format. The default format will
            be used if `save_format` is not specified.
        :param input_names: Names to be assigned to the input tensors of the model.
        :param output_names: Names to be assigned to the output tensors of the model.
        :param model_args: Tuple of additional positional and keyword arguments
            which are required for the model's forward during export. Should be
            specified in the following format:
                - (a, b, {'x': None, 'y': y}) for positional and keyword arguments.
                - (a, b, {}) for positional arguments only.
                - ({'x': None, 'y': y},) for keyword arguments only.
        """
        self.prepare_for_export()
        backend = get_backend(self.model)
        if backend is BackendType.TENSORFLOW:
            from nncf.tensorflow.exporter import TFExporter

            exporter = TFExporter(self.model, input_names, output_names, model_args)
        else:
            assert backend is BackendType.TORCH
            from nncf.torch.exporter import PTExporter

            exporter = PTExporter(self.model, input_names, output_names, model_args)
        if save_format is not None:
            exporter.export_model(save_path, save_format)
        else:
            exporter.export_model(save_path)

    def disable_scheduler(self) -> None:
        self._scheduler = CompositeCompressionScheduler()
        for ctrl in self.child_ctrls:
            ctrl.disable_scheduler()
            self._scheduler.add(ctrl.scheduler)

    def get_compression_state(self) -> Dict[str, Any]:
        if self._builder_state is None:
            raise nncf.InternalError("Internal error: builder state is not set for the controller")

        return {self.BUILDER_STATE: self._builder_state, self.CONTROLLER_STATE: self.get_state()}

    def set_builder_state_with_name(self, name: str, builder_state: Dict):
        """
        Sets state of the builder and the corresponding algorithm name. Should be called by the builder to set its
        state and registered algorithm key.

        :param name: algorithm name, the string that was used to register the builder
        :param builder_state: state of the builder
        """
        self._name = name
        self._builder_state = builder_state


class CompositeCompressionAlgorithmBuilder(CompressionAlgorithmBuilder):
    """
    The `CompositeCompressionAlgorithmBuilder` class stores a group of
    `CompressionAlgorithmBuilder` instances as a list of children that are
    treated the same way as a single `CompressionAlgorithmBuilder` instance.
    """

    def __init__(self, config: NNCFConfig, should_init: bool = True):
        """
        Initializes internal state of the composite compression algorithm builder

        :param config: The top-level NNCFConfig object (i.e. parsed from a .json and extended with
            all necessary objects required for compression such as initialization data loaders).
        :param should_init: If False, trainable parameter initialization will be
            skipped during building.
        """
        self._config = config
        self.should_init = should_init
        self._child_builders = []

    def _get_algo_specific_config_section(self) -> Dict:
        return {}

    @property
    def child_builders(self) -> List[CompressionAlgorithmBuilder]:
        return self._child_builders

    def load_state(self, state: Dict[str, Dict]) -> None:
        """
        Loads the compression builder state of children

        :param state: Output of `get_state()` method.
        """
        for builder in self.child_builders:
            builder.load_state(state)

    def get_state(self) -> Dict[str, Dict]:
        """
        Returns the composite compression builder state. This state contains
        the state of all children.

        :return: The composite compression builder state.
        """
        result = {}
        for builder in self.child_builders:
            result.update(builder.get_state())
        return result



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import IntEnum
from typing import Any, Dict

from nncf.common.stateful_classes_registry import CommonStatefulClassesRegistry


class TransformationPriority(IntEnum):
    """
    Defines priorities for compression and service operations that are
    added as modifications to the original model graph in order to turn it into a
    'compressed' model, i.e. a model that supports compression-aware training and
    export. Certain compression algorithms may need to act on one and the same
    spot in the model's control flow graph, and proper priority-based ordering is
    required for functionally correct results that are invariant w.r.t. compression
    algorithm application ordering.

    Rationales:
    * quantization should occur after sparsification/pruning, otherwise the dependency
    of sparsity level on the value threshold becomes discontinuous which impacts the
    corresponding algorithms.
    """

    DEFAULT_PRIORITY = 0
    FP32_TENSOR_STATISTICS_OBSERVATION = 1
    PRUNING_PRIORITY = 2
    SPARSIFICATION_PRIORITY = 3
    QUANTIZATION_PRIORITY = 11


TARGET_TYPE_STATE_ATTR = "name"


class TargetType(IntEnum):
    """
    Describes the types of locations in the model that can be modified using NNCF
    in order to create a compressed model.

    `LAYER` - a location corresponding directly to an existing layer in the model
    `BEFORE_LAYER` - a location before the associated model layer,
                     implemented by inserting an additional layer in the TF model
    `AFTER_LAYER` - a location after the associated model layer,
                    implemented by inserting an additional layer in the TF model
    `PRE_LAYER_OPERATION` - a location before the associated PT-module or TF-layer
                            execution, for which the local attributes of said
                            PT-module or TF-layer are accessible
    `POST_LAYER_OPERATION` - a location after the associated PT-module or TF-layer
                             execution, for which the local attributes of said
                             PT-module or TF-layer are accessible
    `OPERATION_WITH_WEIGHTS` - same as PRE_LAYER_OPERATION, but targets weights
                               of the layer/module specifically
    `OPERATOR_PRE_HOOK` - a location before a function call in PT without access to
                          specific module attributes - N/A in TF
    `OPERATOR_POST_HOOK` - a location after a function call in PT without access to
                           specific module attributes - N/A in TF

    Notes: Adding operations to a PT-module or TF-layer implemented by wrapping
    the original PT-module or TF-layer and registering operations that are executed
    before/after calling the original PT-module or TF-layer according to the
    registration location:`PRE_LAYER_OPERATION`, `POST_LAYER_OPERATION` and
    `OPERATION_WITH_WEIGHTS`.
    """

    LAYER = 0
    BEFORE_LAYER = 1
    AFTER_LAYER = 2
    PRE_LAYER_OPERATION = 3
    POST_LAYER_OPERATION = 4
    OPERATION_WITH_WEIGHTS = 5
    OPERATOR_PRE_HOOK = 6
    OPERATOR_POST_HOOK = 7

    def get_state(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.

        :return: state of the object
        """
        return {TARGET_TYPE_STATE_ATTR: self.name}

    @classmethod
    def from_state(cls, state: Dict[str, Any]) -> "TargetType":
        """
        Creates the object from its state.

        :param state: Output of `get_state()` method.
        """
        return TargetType[state[TARGET_TYPE_STATE_ATTR]]


class TransformationType(IntEnum):
    """
    Defines the types of transformations that can be applied to a location in the control
     flow graph of the model.
    `TransformationType` defines *what* to do, while `TargetType` more concerns itself with
    *where* to do it.
    """

    INSERT = 0
    MULTI_INSERT = 1
    REMOVE = 2
    CHANGE = 3
    EXTRACT = 4


class TargetPointStateNames:
    TARGET_TYPE = "target_type"


@CommonStatefulClassesRegistry.register()
class TargetPoint:
    """
    The base class for all target points.

    A target point is an object or spot in the model graph. It can be a layer,
    weights, position before or after layer and etc.

    For example, the transformation commands use `TargetPoint` to specify
    the target point in the model graph to which the transformation command
    will be applied.
    """

    _state_names = TargetPointStateNames

    def __init__(self, target_type: TargetType):
        """
        Constructor.

        :param target_type: Type of the target point.
        """
        self._target_type = target_type

    @property
    def type(self) -> TargetType:
        return self._target_type

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, TargetPoint) and self.type == other.type

    def __str__(self) -> str:
        return str(self.type)

    def __hash__(self) -> int:
        return hash(str(self))

    def get_state(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.

        :return: state of the object
        """
        return {self._state_names.TARGET_TYPE: self._target_type.get_state()}

    def is_weight_target_point(self) -> bool:
        return self._target_type == TargetType.OPERATION_WITH_WEIGHTS

    @classmethod
    def from_state(cls, state: Dict[str, Any]) -> "TargetPoint":
        """
        Creates the object from its state.

        :param state: Output of `get_state()` method.
        """
        kwargs = {cls._state_names.TARGET_TYPE: TargetType.from_state(state[cls._state_names.TARGET_TYPE])}
        return cls(**kwargs)


class Command:
    """
    The base class for non-target transformation commands.
    """

    def __init__(self, command_type: TransformationType):
        """
        Initializes Command

        :param command_type: The TransformationType of the non-target transformation command.
        """
        self._command_type = command_type

    @property
    def type(self) -> TransformationType:
        return self._command_type


class TransformationCommand(Command):
    """
    The base class for all transformation commands.
    """

    def __init__(self, command_type: TransformationType, target_point: TargetPoint):
        """
        Constructor.

        :param command_type: Type of the transformation command.
        :param target_point: Target point, the object or spot in the model graph
            to which the transformation command will be applied.
        """
        super().__init__(command_type)
        self._target_point = target_point

    @property
    def target_point(self) -> TargetPoint:
        return self._target_point



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from typing import Iterable, List, Optional, Tuple, TypeVar

import nncf
from nncf.common.factory import NNCFGraphFactory
from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph.utils import get_number_of_quantized_ops
from nncf.common.logging import nncf_logger
from nncf.common.quantization.quantizer_removal import revert_operations_to_floating_point_precision
from nncf.common.utils.backend import BackendType
from nncf.common.utils.backend import get_backend
from nncf.common.utils.os import get_available_cpu_count
from nncf.common.utils.os import get_available_memory_amount
from nncf.data.dataset import Dataset
from nncf.parameters import DropType
from nncf.quantization.advanced_parameters import RestoreMode
from nncf.quantization.algorithms.accuracy_control.backend import AccuracyControlAlgoBackend
from nncf.quantization.algorithms.accuracy_control.evaluator import Evaluator
from nncf.quantization.algorithms.accuracy_control.evaluator import MetricResults
from nncf.quantization.algorithms.accuracy_control.ranker import Ranker

TModel = TypeVar("TModel")
TTensor = TypeVar("TTensor")
PREPARATION_MODEL_THRESHOLD = 1
OVERHEAD_COEFFICIENT = 2
MEMORY_INCREASE_COEFFICIENT = 2


def get_algo_backend(backend: BackendType) -> AccuracyControlAlgoBackend:
    """
    Returns backend for accuracy control algorithm.

    :param backend: Backend.
    :return: The backend for accuracy control algorithm.
    """
    if backend == BackendType.OPENVINO:
        from nncf.quantization.algorithms.accuracy_control.openvino_backend import OVAccuracyControlAlgoBackend

        return OVAccuracyControlAlgoBackend()

    if backend == BackendType.ONNX:
        from nncf.quantization.algorithms.accuracy_control.onnx_backend import ONNXAccuracyControlAlgoBackend

        return ONNXAccuracyControlAlgoBackend()

    raise nncf.UnsupportedBackendError(
        f"Cannot create the backend for the accuracy control algorithm because {backend} is not supported."
    )


def _create_message(nodes: Iterable[NNCFNode]) -> str:
    names = [f"\t{x.node_name}" for x in nodes]
    return "\n".join(names)


def calculate_accuracy_drop(
    initial_metric: float, quantized_metric: float, max_drop: float, drop_type: DropType
) -> Tuple[bool, Optional[float]]:
    """
    Calculates accuracy drop and termination boolean flag.

    :param initial_metric: Metric value for initial model.
    :param quantized_metric: Metric value for quantized model.
    :param max_drop: Maximum accuracy drop that should be achieved.
    :param drop_type: Accuracy drop type.
    :return: A tuple (should_terminate, accuracy_drop) where:
        - should_terminate: Whether the algorithm should terminate or not.
        - accuracy_drop: Accuracy drop value.
    """
    should_terminate = None
    accuracy_drop = None

    if quantized_metric >= initial_metric:
        drop_values_by_drop_type = {
            DropType.RELATIVE: None,
            DropType.ABSOLUTE: initial_metric - quantized_metric,
        }
        accuracy_drop = drop_values_by_drop_type[drop_type]
        should_terminate = True
    else:
        drop_values_by_drop_type = {
            DropType.RELATIVE: abs(1 - quantized_metric / initial_metric),
            DropType.ABSOLUTE: initial_metric - quantized_metric,
        }
        accuracy_drop = drop_values_by_drop_type[drop_type]
        should_terminate = accuracy_drop <= max_drop

    return should_terminate, accuracy_drop


class QuantizationAccuracyRestorerReport:
    """
    Contains execution information about accuracy-aware algorithm.

    :param removed_groups: All groups of quantizers which were removed.
    :param removed_all: True if all quantizers were removed, False otherwise.
    :param reached_required_drop: True if the required accuracy drop was reached, False otherwise.
    :param num_quantized_operations: Number of quantized operations in the model.
    :param num_iterations: Number of iterations performed.
    """

    def __init__(self):
        self.removed_groups = []
        self.removed_all = False
        self.reached_required_drop = False
        self.num_quantized_operations = None
        self.num_iterations = None

    @property
    def removed_quantizers(self) -> List[NNCFNode]:
        """
        Returns all removed quantizers during accuracy-aware algorithm.
        """
        quantizers = []
        for group in self.removed_groups:
            quantizers.extend(group.quantizers)
        return quantizers

    @property
    def reverted_operations(self) -> List[NNCFNode]:
        """
        Returns all operations which were reverted to original precision
        during accuracy-aware algorithm.
        """
        operations = []
        for group in self.removed_groups:
            operations.extend(group.operations)
        return operations


class QuantizationAccuracyRestorer:
    """
    Implementation of the accuracy-aware loop.
    """

    def __init__(
        self,
        ranking_subset_size: int = 300,
        max_num_iterations: int = sys.maxsize,
        max_drop: float = 0.01,
        drop_type: DropType = DropType.ABSOLUTE,
        num_ranking_workers: Optional[int] = None,
        restore_mode: RestoreMode = RestoreMode.ACTIVATIONS_AND_WEIGHTS,
    ):
        """
        :param ranking_subset_size: The number of data items that will be selected from
            the dataset to rank groups of quantizers.
        :param max_num_iterations: A maximal number of iterations.
        :param max_drop: The maximum accuracy drop that should be achieved.
        :param drop_type: The accuracy drop type, which determines how the maximum
            accuracy drop between the original model and the compressed model is
            calculated.
        :param num_ranking_workers: The number of parallel workers that are used to rank
            quantization operations.
        :param restore_mode: Restore mode.
        """
        self.ranking_subset_size = ranking_subset_size
        self.max_num_iterations = max_num_iterations
        self.max_drop = max_drop
        self.drop_type = drop_type
        self.num_ranking_workers = num_ranking_workers
        self.restore_mode = restore_mode

    def apply(
        self,
        initial_model: TModel,
        initial_metric_results: MetricResults,
        quantized_model: TModel,
        quantized_metric_results: MetricResults,
        validation_dataset: Dataset,
        validation_dataset_size: int,
        evaluator: Evaluator,
    ) -> TModel:
        """
        Restores the accuracy of the quantized model by removing the groups of quantizers
        that contribute the most to the drop in accuracy.

        :param initial_model: Initial model (not quantized).
        :param initial_metric_results: Initial model metrics.
        :param quantized_model: Quantized model.
        :param quantized_metric_results: Quantized model metrics.
        :param validation_dataset: A dataset for the validation process.
        :param validation_dataset_size: Validation dataset size.
        :param evaluator: The instance of `Evaluator` to validate model and collect values
            for each item from dataset.
        :return: The quantized model whose metric `final_metric` is satisfied
            the maximum accuracy drop condition.
        """
        algo_backend = get_algo_backend(get_backend(initial_model))

        should_terminate, accuracy_drop = calculate_accuracy_drop(
            initial_metric_results.metric_value, quantized_metric_results.metric_value, self.max_drop, self.drop_type
        )

        if should_terminate:
            QuantizationAccuracyRestorer._print_completion_message(accuracy_drop, self.drop_type)
            return quantized_model

        nncf_logger.info(f"Accuracy drop: {accuracy_drop} ({self.drop_type})")

        # Accuracy drop is greater than the maximum drop so we need to restore accuracy
        return self._apply(
            initial_model,
            initial_metric_results,
            quantized_model,
            quantized_metric_results,
            validation_dataset,
            validation_dataset_size,
            evaluator,
            accuracy_drop,
            algo_backend,
        )

    def _apply(
        self,
        initial_model: TModel,
        initial_metric_results: MetricResults,
        quantized_model: TModel,
        quantized_metric_results: MetricResults,
        validation_dataset: Dataset,
        validation_dataset_size: int,
        evaluator: Evaluator,
        accuracy_drop: float,
        algo_backend: AccuracyControlAlgoBackend,
    ) -> TModel:
        """
        An internal function that implements an iterative approach to restoring the accuracy of
        the quantized model by removing the groups of quantizers that contribute the most to
        the drop in accuracy.

        :param initial_model: Initial model (not quantized).
        :param initial_metric_results: Initial model metrics.
        :param quantized_model: Quantized model.
        :param quantized_metric_results: Quantized model metrics.
        :param validation_dataset: A dataset for the validation process.
        :param validation_dataset_size: Validation dataset size.
        :param evaluator: The instance of `Evaluator` to validate model and collect values
            for each item from dataset.
        :param accuracy_drop: Accuracy drop between initial and quantized models.
        :param algo_backend: The `AccuracyControlAlgoBackend` algo backend.
        :return: The quantized model whose metric `final_metric` is satisfied
            the maximum accuracy drop condition.
        """
        initial_model_graph = NNCFGraphFactory.create(initial_model)
        quantized_model_graph = NNCFGraphFactory.create(quantized_model)

        # Collect original biases and weights because these values are
        # required to undo bias correction and weight correction.
        # Store this data inside the `node.attributes` dictionary.
        # This data will be used in the `revert_operations_to_floating_point_precision()` method.
        QuantizationAccuracyRestorer._collect_original_biases_and_weights(
            initial_model_graph, quantized_model_graph, initial_model, algo_backend
        )

        # Show the number of quantized operations in the model.
        report = QuantizationAccuracyRestorerReport()
        report.num_quantized_operations = get_number_of_quantized_ops(
            quantized_model_graph, algo_backend.get_quantizer_metatypes(), algo_backend.get_quantizable_metatypes()
        )
        nncf_logger.info(f"Total number of quantized operations in the model: {report.num_quantized_operations}")

        # Calculate number of parallel processes for Ranker
        num_ranking_workers = self.num_ranking_workers
        if num_ranking_workers is None:
            model_size = algo_backend.get_model_size(quantized_model)
            num_ranking_workers = self._calculate_number_ranker_workers(
                model_size,
                quantized_metric_results.preparation_time,
                quantized_metric_results.validation_time,
                validation_dataset_size,
            )

        nncf_logger.info(f"Number of parallel workers to rank quantized operations: {num_ranking_workers}")

        ranker = Ranker(self.ranking_subset_size, validation_dataset, algo_backend, evaluator, num_ranking_workers)
        groups_to_rank = ranker.find_groups_of_quantizers_to_rank(quantized_model_graph)
        ranked_groups = ranker.rank_groups_of_quantizers(
            groups_to_rank,
            quantized_model,
            quantized_model_graph,
            initial_metric_results.values_for_each_item,
            quantized_metric_results.values_for_each_item,
        )

        previous_model = quantized_model
        previous_approximate_values_for_each_item = quantized_metric_results.values_for_each_item
        previous_accuracy_drop = accuracy_drop
        current_model = None
        current_approximate_values_for_each_item = None
        current_accuracy_drop = None
        is_step_back = True

        nncf_logger.info("Changing the scope of quantizer nodes was started")
        for iteration in range(self.max_num_iterations):
            if current_model is not None:
                previous_model = current_model

            # greedy removal of the FQ node with the highest importance score
            current_group = ranked_groups.pop()
            current_model = revert_operations_to_floating_point_precision(
                current_group.operations,
                current_group.quantizers,
                previous_model,
                quantized_model_graph,
                self.restore_mode,
                algo_backend.get_op_with_weights_metatypes(),
                algo_backend.is_node_with_weight,
                algo_backend.get_weight_tensor_port_ids,
            )
            report.removed_groups.append(current_group)

            nncf_logger.debug(
                f"Removed a block of {len(current_group.quantizers)} quantizers:"
                f"\n{_create_message(current_group.quantizers)}"
            )
            nncf_logger.info(
                f"Reverted {len(current_group.operations)} operations to the floating-point "
                f"precision: \n{_create_message(current_group.operations)}"
            )

            # Calculate drop for new quantization scope.
            current_metric, current_approximate_values_for_each_item = evaluator.validate(
                current_model, validation_dataset
            )

            should_terminate, current_accuracy_drop = calculate_accuracy_drop(
                initial_metric_results.metric_value, current_metric, self.max_drop, self.drop_type
            )

            if not ranked_groups:
                nncf_logger.info(
                    "All layers have been checked and the AccuracyAwareQuantization "
                    "will not be able to achieve the required accuracy drop"
                )
                report.removed_all = True
                break

            # Accuracy was restored to the acceptable drop.
            if should_terminate:
                report.reached_required_drop = True
                QuantizationAccuracyRestorer._print_completion_message(current_accuracy_drop, self.drop_type)
                break

            nncf_logger.info(
                f"Accuracy drop with the new quantization scope is {float(current_accuracy_drop)} ({self.drop_type})"
            )

            # Continue greedy quantizer remove
            if current_accuracy_drop <= previous_accuracy_drop or (
                current_accuracy_drop > previous_accuracy_drop and is_step_back
            ):
                is_step_back = False
                previous_accuracy_drop = current_accuracy_drop
                continue

            if current_accuracy_drop > previous_accuracy_drop:
                current_model = previous_model
                current_approximate_values_for_each_item = previous_approximate_values_for_each_item
                report.removed_groups.pop()
                ranked_groups.append(current_group)
                is_step_back = True

            previous_accuracy_drop = current_accuracy_drop

            nncf_logger.info("Re-calculating ranking scores for remaining groups")
            if current_approximate_values_for_each_item is None:
                current_approximate_values_for_each_item = evaluator.collect_values_for_each_item(
                    current_model, validation_dataset
                )

            ranked_groups = ranker.rank_groups_of_quantizers(
                ranked_groups,
                current_model,
                quantized_model_graph,
                initial_metric_results.values_for_each_item,
                current_approximate_values_for_each_item,
            )

        report.num_iterations = iteration
        QuantizationAccuracyRestorer._print_report(report, self.max_num_iterations)

        return current_model

    def _calculate_number_ranker_workers(
        self,
        model_size: int,
        preparation_time: float,
        validation_time: float,
        validation_dataset_size: int,
    ) -> int:
        """
        Calculate the number of parallel ranker workers

        :param model_size: Target model size.
        :param preparation_time: The time it takes to prepare the model.
        :param validation_time: The time it takes to validate the model.
        :param validation_dataset_size: Validation dataset size.
        :return: The number of parallel ranker workers
        """
        if preparation_time < PREPARATION_MODEL_THRESHOLD:
            return 1

        # Calculate the number of parallel processes needed to override model preparation and
        # metric calculation on the ranking subset
        ranking_time = validation_time * self.ranking_subset_size / validation_dataset_size
        n_workers = max(round((preparation_time / ranking_time + 1) * OVERHEAD_COEFFICIENT), 2)

        # Apply limitation by number of CPU cores
        n_cores = get_available_cpu_count(logical=True)
        n_workers = max(min(n_workers, n_cores // 2), 1)

        # Apply limitation by memory
        ram = get_available_memory_amount()
        n_copies = ram // (model_size * MEMORY_INCREASE_COEFFICIENT)
        n_workers = max(min(n_workers, n_copies - 1), 1)

        return n_workers

    @staticmethod
    def _collect_original_biases_and_weights(
        initial_model_graph: NNCFGraph,
        quantized_model_graph: NNCFGraph,
        initial_model: TModel,
        algo_backend: AccuracyControlAlgoBackend,
    ) -> None:
        """
        Collects initial biases and weights and stores them inside the `node.attributes['original_bias']` and
        `node.attributes['original_weight']` where `node` is a node from `quantized_model_graph`.

        :param initial_model_graph: Graph for initial model.
        :param quantized_model_graph: Graph for quantized model.
        :param initial_model: Initial model.
        :param algo_backend: The `AccuracyControlAlgoBackend` algo backend.
        """
        for node in initial_model_graph.get_all_nodes():
            if algo_backend.is_node_with_bias(node, initial_model_graph):
                node_with_bias = quantized_model_graph.get_node_by_name(node.node_name)
                node_with_bias.attributes["original_bias"] = algo_backend.get_bias_value(
                    node, initial_model_graph, initial_model
                )
            if algo_backend.is_node_with_weight(node):
                node_with_weight = quantized_model_graph.get_node_by_name(node.node_name)
                for port_id in algo_backend.get_weight_tensor_port_ids(node_with_weight):
                    weight = algo_backend.get_weight_value(node, initial_model, port_id)
                    node_with_weight.attributes[f"original_weight.{port_id}"] = weight

    @staticmethod
    def _print_report(report: QuantizationAccuracyRestorerReport, max_num_iterations: int) -> None:
        """
        Shows report.

        :param report: Report.
        :param max_num_iterations: A maximal number of iterations.
        """
        if report.removed_all or not report.reached_required_drop:
            nncf_logger.info("The algorithm could not achieve the required accuracy drop.")

        if report.num_iterations + 1 >= max_num_iterations:
            nncf_logger.info("Maximum number of iteration was reached.")

        if not report.removed_all:
            nncf_logger.debug(f"Quantizers that were removed:\n{_create_message(report.removed_quantizers)}")
            nncf_logger.info(
                f"{len(report.reverted_operations)} out of {report.num_quantized_operations} "
                "were reverted back to the floating-point precision:"
                f"\n{_create_message(report.reverted_operations)}"
            )

    @staticmethod
    def _print_completion_message(accuracy_drop: float, drop_type: DropType) -> None:
        if accuracy_drop is None or accuracy_drop < 0:
            reason = "metric of the quantized model is greater than the metric of the initial model"
        else:
            reason = f"achieved required accuracy drop {float(accuracy_drop)} ({drop_type})"
        nncf_logger.info(f"Algorithm completed: {reason}")



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from nncf.config.definitions import ADAPTIVE_COMPRESSION_LEVEL_TRAINING_MODE_NAME_IN_CONFIG
from nncf.config.definitions import EARLY_EXIT_TRAINING_MODE_NAME_IN_CONFIG
from nncf.config.definitions import ONLINE_DOCS_ROOT
from nncf.config.schemata.basic import NUMBER
from nncf.config.schemata.basic import with_attributes
from nncf.config.schemata.defaults import AA_COMPRESSION_RATE_STEP_REDUCTION_FACTOR
from nncf.config.schemata.defaults import AA_INITIAL_COMPRESSION_RATE_STEP
from nncf.config.schemata.defaults import AA_INITIAL_TRAINING_PHASE_EPOCHS
from nncf.config.schemata.defaults import AA_LR_REDUCTION_FACTOR
from nncf.config.schemata.defaults import AA_MAXIMAL_TOTAL_EPOCHS
from nncf.config.schemata.defaults import AA_MINIMAL_COMPRESSION_RATE_STEP
from nncf.config.schemata.defaults import AA_PATIENCE_EPOCHS

COMMON_AA_PROPERTIES = {
    "maximal_relative_accuracy_degradation": with_attributes(
        NUMBER,
        description="Maximally allowed accuracy degradation of the model "
        "in percent relative to the original model accuracy.",
    ),
    "maximal_absolute_accuracy_degradation": with_attributes(
        NUMBER,
        description="Maximally allowed accuracy degradation of the model "
        "in units of absolute metric of the original model.",
    ),
}

ADAPTIVE_COMPRESSION_LEVEL_TRAINING_SCHEMA = {
    "type": "object",
    "title": ADAPTIVE_COMPRESSION_LEVEL_TRAINING_MODE_NAME_IN_CONFIG,
    "description": f"Adaptive compression level training mode schema. See "
    f"[AdaptiveCompressionLevelTraining.md]"
    f"({ONLINE_DOCS_ROOT}docs/accuracy_aware_model_training/AdaptiveCompressionLevelTraining.md) "
    f"for more general info on this mode.",
    "properties": {
        "mode": {"const": ADAPTIVE_COMPRESSION_LEVEL_TRAINING_MODE_NAME_IN_CONFIG},
        "params": {
            "type": "object",
            "properties": {
                **COMMON_AA_PROPERTIES,
                "initial_training_phase_epochs": with_attributes(
                    NUMBER,
                    description="Number of epochs to fine-tune during the initial "
                    "training phase of the adaptive compression training loop.",
                    default=AA_INITIAL_TRAINING_PHASE_EPOCHS,
                ),
                "initial_compression_rate_step": with_attributes(
                    NUMBER,
                    description="Initial value for the compression rate increase/decrease "
                    "training phase of the compression training loop.",
                    default=AA_INITIAL_COMPRESSION_RATE_STEP,
                ),
                "compression_rate_step_reduction_factor": with_attributes(
                    NUMBER,
                    description="Factor used to reduce the compression rate change step "
                    "in the adaptive compression training loop.",
                    default=AA_COMPRESSION_RATE_STEP_REDUCTION_FACTOR,
                ),
                "lr_reduction_factor": with_attributes(
                    NUMBER,
                    description="Factor used to reduce the learning rate after compression rate step is reduced",
                    default=AA_LR_REDUCTION_FACTOR,
                ),
                "minimal_compression_rate_step": with_attributes(
                    NUMBER,
                    description="The minimal compression rate change step value "
                    "after which the training loop is terminated.",
                    default=AA_MINIMAL_COMPRESSION_RATE_STEP,
                ),
                "patience_epochs": with_attributes(
                    NUMBER,
                    description="The number of epochs to fine-tune the model "
                    "for a given compression rate after the initial "
                    "training phase of the training loop.",
                    default=AA_PATIENCE_EPOCHS,
                ),
                "maximal_total_epochs": with_attributes(
                    NUMBER,
                    description="The maximal total fine-tuning epoch count. If the epoch "
                    "counter reaches this number, the fine-tuning process will "
                    "stop and the model with the largest compression rate "
                    "will be returned.",
                    default=AA_MAXIMAL_TOTAL_EPOCHS,
                ),
            },
            "oneOf": [
                {"required": ["maximal_relative_accuracy_degradation"]},
                {"required": ["maximal_absolute_accuracy_degradation"]},
            ],
            "required": ["initial_training_phase_epochs", "patience_epochs"],
            "additionalProperties": False,
        },
    },
    "required": ["mode", "params"],
    "additionalProperties": False,
}
EARLY_EXIT_TRAINING_SCHEMA = {
    "type": "object",
    "title": EARLY_EXIT_TRAINING_MODE_NAME_IN_CONFIG,
    "description": f"Early exit mode schema. See "
    f"[EarlyExitTraining.md]"
    f"({ONLINE_DOCS_ROOT}docs/accuracy_aware_model_training/EarlyExitTraining.md) for "
    f"more general info on this mode.",
    "properties": {
        "mode": {"const": EARLY_EXIT_TRAINING_MODE_NAME_IN_CONFIG},
        "params": {
            "type": "object",
            "properties": {
                **COMMON_AA_PROPERTIES,
                "maximal_total_epochs": with_attributes(
                    NUMBER,
                    description="The maximal total fine-tuning epoch count. If the accuracy criteria "
                    "wouldn't reach during fine-tuning, the most accurate model "
                    "will be returned.",
                    default=AA_MAXIMAL_TOTAL_EPOCHS,
                ),
            },
            "oneOf": [
                {"required": ["maximal_relative_accuracy_degradation"]},
                {"required": ["maximal_absolute_accuracy_degradation"]},
            ],
            "required": ["maximal_total_epochs"],
            "additionalProperties": False,
        },
    },
    "required": ["mode", "params"],
    "additionalProperties": False,
}
ACCURACY_AWARE_TRAINING_SCHEMA = {
    "type": "object",
    "oneOf": [EARLY_EXIT_TRAINING_SCHEMA, ADAPTIVE_COMPRESSION_LEVEL_TRAINING_SCHEMA],
}
ACCURACY_AWARE_MODES_VS_SCHEMA = {
    ADAPTIVE_COMPRESSION_LEVEL_TRAINING_MODE_NAME_IN_CONFIG: ADAPTIVE_COMPRESSION_LEVEL_TRAINING_SCHEMA,
    EARLY_EXIT_TRAINING_MODE_NAME_IN_CONFIG: EARLY_EXIT_TRAINING_SCHEMA,
}



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Optional, Set, Tuple

import onnx

from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph.transformations.commands import TargetType
from nncf.experimental.common.tensor_statistics.collectors import TensorCollector
from nncf.onnx.graph.model_utils import remove_fq_from_inputs
from nncf.onnx.graph.node_utils import get_bias_value
from nncf.onnx.graph.node_utils import is_any_weight_quantized
from nncf.onnx.graph.node_utils import is_node_with_bias
from nncf.onnx.graph.onnx_helper import get_name_to_node_map
from nncf.onnx.graph.transformations.command_creation import create_bias_correction_command
from nncf.onnx.graph.transformations.commands import ONNXInitializerUpdateCommand
from nncf.onnx.graph.transformations.commands import ONNXModelExtractionCommand
from nncf.onnx.graph.transformations.commands import ONNXOutputInsertionCommand
from nncf.onnx.graph.transformations.commands import ONNXTargetPoint
from nncf.onnx.statistics.collectors import get_mean_statistic_collector
from nncf.onnx.statistics.collectors import get_raw_stat_collector
from nncf.quantization.algorithms.bias_correction.backend import BiasCorrectionAlgoBackend
from nncf.tensor import Tensor


class ONNXBiasCorrectionAlgoBackend(BiasCorrectionAlgoBackend):

    @staticmethod
    def target_point(target_type: TargetType, target_node_name: str, port_id: int) -> ONNXTargetPoint:
        return ONNXTargetPoint(target_type, target_node_name, port_id)

    @staticmethod
    def create_bias_correction_command(
        node: NNCFNode, bias_value: Tensor, nncf_graph: NNCFGraph
    ) -> ONNXInitializerUpdateCommand:
        return create_bias_correction_command(node, bias_value.data)

    @staticmethod
    def model_extraction_command(
        input_ids: Set[Tuple[str, int]], output_ids: Set[Tuple[str, int]]
    ) -> ONNXModelExtractionCommand:
        return ONNXModelExtractionCommand(input_ids, output_ids)

    @staticmethod
    def output_insertion_command(nncf_graph: NNCFGraph, target_point: ONNXTargetPoint) -> ONNXOutputInsertionCommand:
        nncf_input_node_next_nodes = {}
        for input_node in nncf_graph.get_input_nodes():
            next_nodes = nncf_graph.get_next_nodes(input_node)
            nncf_input_node_next_nodes[input_node.node_name] = [node.node_name for node in next_nodes]
        return ONNXOutputInsertionCommand(target_point, nncf_input_node_next_nodes)

    @staticmethod
    def mean_statistic_collector(
        channel_axis: int,
        inplace: bool,
        num_samples: Optional[int] = None,
        window_size: Optional[int] = None,
    ) -> TensorCollector:
        return get_mean_statistic_collector(num_samples, channel_axis, window_size, inplace)

    @staticmethod
    def raw_statistic_collector(num_samples: int = None) -> TensorCollector:
        return get_raw_stat_collector(num_samples)

    @staticmethod
    def process_model_output(raw_data: Dict, output_name: str) -> Tensor:
        return Tensor(raw_data[output_name])

    @staticmethod
    def get_activation_port_id(node: NNCFNode, nncf_graph: NNCFGraph) -> Tuple[int, int]:
        return 0

    @staticmethod
    def get_bias_value(node: NNCFNode, model: onnx.ModelProto, nncf_graph: NNCFGraph) -> Tensor:
        return Tensor(get_bias_value(node, model))

    @staticmethod
    def get_input_name(model: onnx.ModelProto, node_name: str, input_port_id: int) -> str:
        node_mapping = get_name_to_node_map(model)
        return node_mapping[node_name].input[input_port_id]

    @staticmethod
    def get_output_name(model: onnx.ModelProto, node_name: str, output_port_id: int) -> str:
        node_mapping = get_name_to_node_map(model)
        return node_mapping[node_name].output[output_port_id]

    @staticmethod
    def is_quantized_weights(node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
        return is_any_weight_quantized(node, nncf_graph)

    @staticmethod
    def is_node_with_bias(node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
        return is_node_with_bias(node)

    @staticmethod
    def remove_fq_from_inputs(model: onnx.ModelProto, nncf_graph: NNCFGraph) -> onnx.ModelProto:
        return remove_fq_from_inputs(model, nncf_graph)

    @staticmethod
    def get_port_id(target_point: ONNXTargetPoint) -> int:
        return target_point.port_id



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from enum import Enum


class NamespaceTarget(Enum):
    """
    NamespaceTarget stores modules from which patched operators were obtained.
    """

    TORCH_NN_FUNCTIONAL = "torch.nn.functional"
    TORCH_TENSOR = "torch.tensor"
    TORCH_NN_PARAMETER = "torch.nn.parameter"
    TORCH = "torch"
    ATEN = "aten"
    EXTERNAL = "external_function"


class PatchedOperatorInfo:
    def __init__(self, name: str, operator_namespace: NamespaceTarget, skip_trace: bool = False):
        """
        Information about patched operator.
        :param name: Operator name
        :param operator_namespace: Python module, from which operator was gotten.
        :param skip_trace: If it is set to True, the both operator and its internal calls
         to otherwise traced functions do not appear into the model graph.
        """
        self.name = name
        self.operator_namespace = operator_namespace
        self.skip_trace = skip_trace



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional, Tuple

import numpy as np

from nncf.common.graph.transformations.commands import Command
from nncf.common.graph.transformations.commands import TargetPoint
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.graph.transformations.commands import TransformationCommand
from nncf.common.graph.transformations.commands import TransformationType
from nncf.onnx.quantization.quantizer_parameters import ONNXQuantizerLayerParameters


class ONNXTargetPoint(TargetPoint):
    def __init__(self, target_type: TargetType, target_node_name: str, port_id: Optional[int] = None):
        """
        Constructor.

        :param target_type: Target type of the target point.
        :param target_node_name: ONNX node name.
        :param port_id: Number of port id in case target_type is
            TargetType.PRE_LAYER_OPERATION or TargetType.POST_LAYER_OPERATION.
        """
        super().__init__(target_type)
        self.target_node_name = target_node_name
        self.port_id = port_id

    def __eq__(self, other: "ONNXTargetPoint") -> bool:
        return (
            isinstance(other, ONNXTargetPoint)
            and self.type == other.type
            and self.target_node_name == other.target_node_name
            and self.port_id == other.port_id
        )

    def __hash__(self) -> int:
        return hash((self.target_node_name, self.port_id, self._target_type))


class ONNXInsertionCommand(TransformationCommand):
    def __init__(self, target_point: ONNXTargetPoint, input_edges_mapping: Dict[str, Tuple[str, int]]):
        super().__init__(TransformationType.INSERT, target_point)
        # Insertion command could be applied to NNCF input nodes, e.g.
        # quantizers will be tied with POST OP of NNCF input nodes.
        # To get the ONNX edge to apply a command,
        # need to keep the mapping NNCF input nodes to the following ONNX nodes.
        self.input_edges_mapping = input_edges_mapping


class ONNXQuantizerInsertionCommand(ONNXInsertionCommand):
    def __init__(
        self,
        target_point: ONNXTargetPoint,
        nncf_input_node_next_onnx_nodes: Dict[str, List[str]],
        quantizer_parameters: ONNXQuantizerLayerParameters,
    ):
        super().__init__(target_point, nncf_input_node_next_onnx_nodes)
        self.quantizer_parameters = quantizer_parameters


class ONNXOutputInsertionCommand(ONNXInsertionCommand):
    pass


class ONNXInitializerUpdateCommand(TransformationCommand):
    """
    Update initializer in the value.
    """

    def __init__(self, target_point: ONNXTargetPoint, new_value: np.ndarray):
        """
        :param target_point: Target point.
        :param new_value: New value for initializer.
        """
        super().__init__(TransformationType.CHANGE, target_point)
        self.new_value = new_value


class ONNXModelExtractionCommand(Command):
    """
    Extracts sub-graph based on the sub-model input and output names.
    """

    def __init__(self, input_ids: List[Tuple[str, int]], output_ids: List[Tuple[str, int]]):
        """
        :param input_ids: List of the input IDs: pairs of node names and correspondent input port ids.
            Each pair denotes the sub-graph beginning.
        :param output_ids: List of the output IDs: pairs of node names and correspondent output port ids.
            Each pair denotes the sub-graph ending.
        """
        super().__init__(TransformationType.EXTRACT)
        self.input_ids = input_ids
        self.output_ids = output_ids


class ONNXQDQNodeRemovingCommand(TransformationCommand):
    """
    Removes Quantizer or Dequantizer nodes from the model.
    """

    def __init__(self, target_point: ONNXTargetPoint):
        """
        :param target_point: The TargetPoint instance for the layer that contains information for removing.
        """
        super().__init__(TransformationType.REMOVE, target_point)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TypeVar

from nncf.common.graph.transformations.layout import TransformationLayout

TModel = TypeVar("TModel")


class ModelTransformer:
    """
    Applies transformations to the model.
    """

    def __init__(self, model: TModel):
        """
        Initializes Model Transformer.

        :param model: The model to be transformed.
        """
        self._model = model

    def transform(self, transformation_layout: TransformationLayout) -> TModel:  # type:ignore
        """
        Applies transformations to the model.

        :param transformation_layout: An instance of `TransformationLayout` that
            includes a list of transformations to be applied to the model.
        :return: The transformed model.
        """
        raise NotImplementedError()
